# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-05-13 11:15
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("contentstore", "0002_messageset_content_type")]

    operations = [
        migrations.AlterField(
            model_name="messageset",
            name="short_name",
            field=models.CharField(
                max_length=100, unique=True, verbose_name="Short name"
            ),
        )
    ]



import json
import os
from datetime import datetime
from unittest.mock import patch

import pytz
import responses
from django.conf import settings
from django.contrib.auth.models import User
from django.test import TestCase
from django.urls import reverse
from rest_framework import status
from rest_framework.authtoken.models import Token
from rest_framework.test import APIClient

from seed_stage_based_messaging import test_utils as utils

from ..models import BinaryContent, Message, MessageSet, Schedule


class APITestCase(TestCase):
    def setUp(self):
        self.client = APIClient()
        utils.disable_signals()

    def tearDown(self):
        utils.enable_signals()


class MessageSetTestMixin:
    def make_schedule(self):
        # Create hourly schedule
        schedule_data = {"minute": "0"}
        return Schedule.objects.create(**schedule_data)

    def make_messageset(
        self,
        short_name="messageset_one",
        notes=None,
        next_set=None,
        schedule=None,
        channel=None,
    ):
        if schedule is None:
            schedule = self.make_schedule()
        messageset_data = {
            "short_name": short_name,
            "notes": notes,
            "next_set": next_set,
            "default_schedule": schedule,
            "channel": channel,
        }
        return MessageSet.objects.create(**messageset_data)


class AuthenticatedAPITestCase(MessageSetTestMixin, APITestCase):
    def setUp(self):
        super(AuthenticatedAPITestCase, self).setUp()

        self.username = "testuser"
        self.password = "testpass"
        self.user = User.objects.create_user(
            self.username, "testuser@example.com", self.password
        )
        token = Token.objects.create(user=self.user)
        self.token = token.key
        self.client.credentials(HTTP_AUTHORIZATION="Token " + self.token)


class TestLogin(AuthenticatedAPITestCase):
    def test_login(self):
        # Setup
        post_auth = {"username": "testuser", "password": "testpass"}
        # Execute
        request = self.client.post("/api/token-auth/", post_auth)
        token = request.data.get("token", None)
        # Check
        self.assertIsNotNone(
            token, "Could not receive authentication token on login post."
        )
        self.assertEqual(
            request.status_code,
            200,
            "Status code on /api/token-auth was %s (should be 200)."
            % request.status_code,
        )


class TestContentStoreApi(AuthenticatedAPITestCase):

    # Schedule testing
    def test_read_schedule(self):
        # Setup
        existing = self.make_schedule()
        # Execute
        response = self.client.get(
            "/api/v1/schedule/%s/" % existing.id, content_type="application/json"
        )
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["hour"], "*")
        self.assertEqual(response.data["minute"], "0")
        d = Schedule.objects.last()
        self.assertEqual(d.cron_string, "0 * * * *")

    def test_filter_schedule(self):
        # Setup
        existing = self.make_schedule()
        # Execute
        response = self.client.get(
            "/api/v1/schedule/",
            {"cron_string": "* 1 * * *"},
            content_type="application/json",
        )
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["results"][0]["id"], existing.id)

    def test_list_schedule(self):
        # Setup
        schedules = []
        for i in range(3):
            schedules.append(self.make_schedule())
        # Execute
        response = self.client.get("/api/v1/schedule/", content_type="application/json")

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # Check first page
        body = response.json()
        self.assertEqual(len(body["results"]), 2)
        self.assertEqual(body["results"][0]["id"], schedules[0].id)
        self.assertEqual(body["results"][1]["id"], schedules[1].id)
        self.assertIsNone(body["previous"])
        self.assertIsNotNone(body["next"])

        # Check next page
        body = self.client.get(body["next"]).json()
        self.assertEqual(len(body["results"]), 1)
        self.assertEqual(body["results"][0]["id"], schedules[2].id)
        self.assertIsNotNone(body["previous"])
        self.assertIsNone(body["next"])

        # Check previous page
        body = self.client.get(body["previous"]).json()
        self.assertEqual(len(body["results"]), 2)
        self.assertEqual(body["results"][0]["id"], schedules[0].id)
        self.assertEqual(body["results"][1]["id"], schedules[1].id)
        self.assertIsNone(body["previous"])
        self.assertIsNotNone(body["next"])

    # MessageSet testing
    def test_read_messageset(self):
        # Setup
        schedule = self.make_schedule()
        messageset = self.make_messageset(schedule=schedule)
        # Execute
        response = self.client.get(
            "/api/v1/messageset/%s/" % messageset.id, content_type="application/json"
        )
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        d = MessageSet.objects.last()
        self.assertIsNotNone(d.id)
        self.assertEqual(d.short_name, "messageset_one")
        self.assertEqual(d.notes, None)
        self.assertEqual(d.next_set, None)
        self.assertEqual(d.default_schedule, schedule)
        self.assertEqual(d.content_type, "text")

    def test_create_messageset(self):
        # Setup
        schedule = self.make_schedule()
        messageset_data = {
            "short_name": "messageset_one_but_very_longname_and_cool_yeah",
            "notes": None,
            "next_set": None,
            "default_schedule": schedule.id,
        }
        # Execute
        response = self.client.post(
            "/api/v1/messageset/",
            json.dumps(messageset_data),
            content_type="application/json",
        )
        # Check
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        d = MessageSet.objects.last()
        self.assertIsNotNone(d.id)
        self.assertEqual(d.short_name, "messageset_one_but_very_longname_and_cool_yeah")
        self.assertEqual(d.notes, None)
        self.assertEqual(d.next_set, None)
        self.assertEqual(d.default_schedule, schedule)
        self.assertEqual(d.content_type, "text")

    def test_list_messagesets(self):
        # Setup
        for i in range(3):
            self.make_messageset(short_name="messageset_%s" % i)
        # Execute
        response = self.client.get(
            "/api/v1/messageset/", content_type="application/json"
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        # Check first page
        body = response.json()
        self.assertEqual(len(body["results"]), 2)
        self.assertEqual(body["results"][0]["short_name"], "messageset_0")
        self.assertEqual(body["results"][1]["short_name"], "messageset_1")
        self.assertIsNone(body["previous"])
        self.assertIsNotNone(body["next"])

        # Check next page
        body = self.client.get(body["next"]).json()
        self.assertEqual(len(body["results"]), 1)
        self.assertEqual(body["results"][0]["short_name"], "messageset_2")
        self.assertIsNotNone(body["previous"])
        self.assertIsNone(body["next"])

        # Check previous page
        body = self.client.get(body["previous"]).json()
        self.assertEqual(len(body["results"]), 2)
        self.assertEqual(body["results"][0]["short_name"], "messageset_0")
        self.assertEqual(body["results"][1]["short_name"], "messageset_1")
        self.assertIsNone(body["previous"])
        self.assertIsNotNone(body["next"])

    def test_filter_messagesets(self):
        # Setup
        self.make_messageset()
        self.make_messageset(short_name="messageset_two")
        # Execute
        response = self.client.get(
            "/api/v1/messageset/",
            {"short_name": "messageset_two"},
            content_type="application/json",
        )
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data["results"]), 1)
        self.assertEqual(response.data["results"][0]["short_name"], "messageset_two")

    def test_create_message(self):
        """
        A POST request should create a message object for a messageset.
        """
        messageset = self.make_messageset()

        data = {
            "messageset": messageset.pk,
            "sequence_number": 1,
            "lang": "en",
            "text_content": "Foo",
        }

        self.client.post(
            reverse("message-list"), json.dumps(data), content_type="application/json"
        )

        [msg] = Message.objects.all()
        self.assertEqual(msg.messageset, messageset)
        self.assertEqual(msg.sequence_number, 1)
        self.assertEqual(msg.lang, "en")
        self.assertEqual(msg.text_content, "Foo")

    def test_create_message_constraint(self):
        """
        When creating a message, if creating a second message with matching
        messageset, sequence_number, and lang fields, it should not be
        created.
        """
        messageset = self.make_messageset()
        Message.objects.create(
            messageset=messageset, sequence_number=1, lang="en", text_content="Foo"
        )

        data = {
            "messageset": messageset.pk,
            "sequence_number": 1,
            "lang": "en",
            "text_content": "Bar",
        }

        response = self.client.post(
            reverse("message-list"), json.dumps(data), content_type="application/json"
        )

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "non_field_errors": [
                    "The fields messageset, sequence_number, lang must make a "
                    "unique set."
                ]
            },
        )
        self.assertEqual(Message.objects.all().count(), 1)

    def make_message(self, messageset, lang, seq=1):
        Message.objects.create(
            messageset=messageset, sequence_number=seq, lang=lang, text_content="Foo"
        )

    def test_list_messages(self):
        messageset = self.make_messageset()
        # Setup
        for i in range(1, 4):
            self.make_message(messageset, "eng_ZA", i)
        # Execute
        response = self.client.get("/api/v1/message/", content_type="application/json")

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        # Check first page
        body = response.json()
        self.assertEqual(len(body["results"]), 2)
        self.assertEqual(body["results"][0]["sequence_number"], 1)
        self.assertEqual(body["results"][1]["sequence_number"], 2)
        self.assertIsNone(body["previous"])
        self.assertIsNotNone(body["next"])

        # Check next page
        body = self.client.get(body["next"]).json()
        self.assertEqual(len(body["results"]), 1)
        self.assertEqual(body["results"][0]["sequence_number"], 3)
        self.assertIsNotNone(body["previous"])
        self.assertIsNone(body["next"])

        # Check previous page
        body = self.client.get(body["previous"]).json()
        self.assertEqual(len(body["results"]), 2)
        self.assertEqual(body["results"][0]["sequence_number"], 1)
        self.assertEqual(body["results"][1]["sequence_number"], 2)
        self.assertIsNone(body["previous"])
        self.assertIsNotNone(body["next"])

    def test_read_messageset_languages(self):
        """
        A GET Request should return all the messagesets with a unique list of
        languages available on each
        """
        # Setup
        messageset = self.make_messageset()
        self.make_message(messageset, "eng")
        self.make_message(messageset, "eng", seq=2)
        self.make_message(messageset, "afr")
        messageset2 = self.make_messageset(short_name="messageset_two")
        self.make_message(messageset2, "eng")
        self.make_message(messageset2, "afr")
        self.make_message(messageset2, "zul")
        self.make_message(messageset2, "zul", seq=2)

        # Execute
        response = self.client.get(
            "/api/v1/messageset_languages/", content_type="application/json"
        )
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            json.loads(response.content.decode("utf-8")),
            {
                str(messageset.id): ["afr", "eng"],
                str(messageset2.id): ["afr", "eng", "zul"],
            },
        )


class TestMessageSet(MessageSetTestMixin, TestCase):
    def setUp(self):
        utils.disable_signals()

    def tearDown(self):
        utils.enable_signals()

    def test_get_all_run_dates(self):
        ms = self.make_messageset()
        for i in range(3):
            Message.objects.create(
                messageset=ms, sequence_number=i, lang="eng_ZA", text_content="Foo"
            )

        start = datetime(2016, 11, 11, 7, 0, tzinfo=pytz.UTC)

        dates = ms.get_all_run_dates(start=start, lang="eng_ZA")

        self.assertEqual(
            dates,
            [
                datetime(2016, 11, 11, 8, 0, tzinfo=pytz.UTC),
                datetime(2016, 11, 11, 9, 0, tzinfo=pytz.UTC),
                datetime(2016, 11, 11, 10, 0, tzinfo=pytz.UTC),
            ],
        )

    def test_get_all_run_dates_none_for_lang(self):
        ms = self.make_messageset()
        for i in range(3):
            Message.objects.create(
                messageset=ms, sequence_number=i, lang="eng_ZA", text_content="Foo"
            )

        start = datetime(2016, 11, 11, 7, 0, tzinfo=pytz.UTC)

        dates = ms.get_all_run_dates(start=start, lang="zul_ZA")

        self.assertEqual(dates, [])

    def test_get_all_run_dates_diff_schedule(self):
        ms = self.make_messageset()
        for i in range(3):
            Message.objects.create(
                messageset=ms, sequence_number=i, lang="eng_ZA", text_content="Foo"
            )

        schedule = Schedule.objects.create(minute="0,30")

        start = datetime(2016, 11, 11, 7, 0, tzinfo=pytz.UTC)

        dates = ms.get_all_run_dates(start=start, lang="eng_ZA", schedule=schedule)

        self.assertEqual(
            dates,
            [
                datetime(2016, 11, 11, 7, 30, tzinfo=pytz.UTC),
                datetime(2016, 11, 11, 8, 0, tzinfo=pytz.UTC),
                datetime(2016, 11, 11, 8, 30, tzinfo=pytz.UTC),
            ],
        )

    def test_get_all_run_dates_diff_initial(self):
        ms = self.make_messageset()
        for i in range(3):
            Message.objects.create(
                messageset=ms, sequence_number=i, lang="eng_ZA", text_content="Foo"
            )

        start = datetime(2016, 11, 11, 7, 0, tzinfo=pytz.UTC)

        dates = ms.get_all_run_dates(
            start=start, lang="eng_ZA", schedule=ms.default_schedule, initial=2
        )

        self.assertEqual(
            dates,
            [
                datetime(2016, 11, 11, 8, 0, tzinfo=pytz.UTC),
                datetime(2016, 11, 11, 9, 0, tzinfo=pytz.UTC),
            ],
        )


class TestSchedule(TestCase):
    def test_cron_string(self):
        schedule = Schedule(
            minute="*", hour="*", day_of_week="1", day_of_month="1", month_of_year="*"
        )
        self.assertEqual(schedule.cron_string, "* * 1 * 1")

        schedule = Schedule(
            minute="0",
            hour="8",
            day_of_week="1, 2, 3",
            day_of_month="*",
            month_of_year="*",
        )
        self.assertEqual(schedule.cron_string, "0 8 * * 1,2,3")

        schedule = Schedule(
            minute="1", hour="2", day_of_week="3", day_of_month="4", month_of_year="5"
        )
        self.assertEqual(schedule.cron_string, "1 2 4 5 3")

    def test_get_run_times_between(self):
        start = datetime(2016, 11, 1, 0, 0, tzinfo=pytz.UTC)
        end = datetime(2016, 11, 30, 23, 59, tzinfo=pytz.UTC)

        # test with every mon, tues, wed in November 2016
        schedule = Schedule(day_of_week="1,2,3", hour="8", minute="0")
        runs = schedule.get_run_times_between(start, end)
        self.assertEqual(len(runs), 14)

        # test with every week day in November
        schedule = Schedule(day_of_week="1,2,3,4,5", hour="8", minute="0")
        runs = schedule.get_run_times_between(start, end)
        self.assertEqual(len(runs), 22)

        # test with every day in November
        schedule = Schedule(day_of_week="*", hour="8", minute="0")
        runs = schedule.get_run_times_between(start, end)
        self.assertEqual(len(runs), 30)

        # test with specific day in November
        schedule = Schedule(day_of_month="21", hour="8", minute="0")
        runs = schedule.get_run_times_between(start, end)
        self.assertEqual(len(runs), 1)


class TestAdmin(MessageSetTestMixin, TestCase):
    def setUp(self):
        utils.disable_signals()
        username = "testuser"
        password = "testpass"
        User.objects.create_superuser(username, "testuser@example.com", password)
        self.client.login(username=username, password=password)
        self.change_url = reverse("admin:contentstore_messageset_changelist")

    def tearDown(self):
        utils.enable_signals()

    def test_clone_action_validation(self):
        message_set = self.make_messageset(short_name="messageset")
        response = self.client.post(
            self.change_url,
            {
                "action": "clone_messageset",
                "do_action": "yes",
                "_selected_action": message_set.pk,
                "short_name": message_set.short_name,
            },
        )
        self.assertContains(response, "A message set already exists with this name.")

    def test_clone_action(self):
        self.assertEqual(MessageSet.objects.count(), 0)
        message_set = self.make_messageset(short_name="messageset")
        self.assertEqual(MessageSet.objects.count(), 1)
        for i in range(10):
            message_set.messages.create(
                sequence_number=i, lang="eng_UK", text_content="message %s" % (i,)
            )

        response = self.client.post(
            self.change_url,
            {
                "action": "clone_messageset",
                "do_action": "yes",
                "_selected_action": message_set.pk,
                "short_name": "new short name!",
            },
        )
        self.assertRedirects(response, self.change_url)
        self.assertEqual(MessageSet.objects.count(), 2)
        clone = MessageSet.objects.all().order_by("-pk").first()
        self.assertEqual(clone.short_name, "new short name!")
        self.assertEqual(clone.messages.count(), message_set.messages.count())
        self.assertNotEqual(clone.pk, message_set.pk)


class TestSyncWelcomeAudio(AuthenticatedAPITestCase):
    def make_binarycontent(self, filename="hello.mp3"):
        data = {"content": filename}
        return BinaryContent.objects.create(**data)

    @responses.activate
    @patch("contentstore.tasks.SyncAudioMessages._get_existing_files")
    @patch("sftpclone.sftpclone.SFTPClone.__init__")
    @patch("sftpclone.sftpclone.SFTPClone.run")
    def test_sync_welcome_audio(self, sftp_run_mock, sftp_mock, get_existing_mock):
        """
        When there is a POST to the sync audio api endpoint, it should sync
        only the missing audio files from the django api container then upload
        it to a sftp folder. The file should be cleaned up after.
        """

        sftp_run_mock.return_value = None
        sftp_mock.return_value = None
        get_existing_mock.return_value = ["other.mp3"]

        self.make_binarycontent()

        responses.add(
            responses.GET,
            "http://example.com/media/hello.mp3",
            "test file content",
            status=200,
            content_type="text/plain",
        )

        response = self.client.post(
            "/api/v1/sync_audio_files/", content_type="application/json"
        )

        self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)

        sftp_mock.assert_called_with(
            "{}/{}/".format(settings.BASE_DIR, settings.MEDIA_ROOT),
            "test:secret@localhost:test_directory",
            port=2222,
            delete=False,
        )

        path = "{}/{}/hello.mp3".format(settings.BASE_DIR, settings.MEDIA_ROOT)
        self.assertFalse(os.path.isfile(path))



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-11-02 09:50
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("contentstore", "0007_auto_20171102_0950"),
        ("subscriptions", "0006_auto_20170727_0820"),
    ]

    operations = [
        migrations.CreateModel(
            name="EstimatedSend",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("send_date", models.DateField()),
                ("estimate_subscriptions", models.IntegerField()),
                ("estimate_identities", models.IntegerField()),
                (
                    "messageset",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="estimates",
                        to="contentstore.MessageSet",
                    ),
                ),
            ],
        ),
        migrations.AlterUniqueTogether(
            name="estimatedsend", unique_together=set([("send_date", "messageset")])
        ),
    ]



from django.conf import settings
from django.core.management.base import BaseCommand
from django.core.validators import URLValidator
from seed_services_client import SchedulerApiClient

from subscriptions.models import Subscription


def url_validator(url_str):
    URLValidator(url_str)
    return url_str


class Command(BaseCommand):
    help = "Removes all duplicate subscriptions"

    DEFAULT_TIME_DELTA = 60
    DEFAULT_FIELDS = [
        "identity",
        "version",
        "messageset",
        "lang",
        "active",
        "completed",
        "schedule",
    ]

    def add_arguments(self, parser):
        parser.add_argument(
            "--time-delta",
            action="store",
            dest="time_delta",
            type=int,
            nargs="?",
            default=self.DEFAULT_TIME_DELTA,
            help="The maximum time difference in seconds between two "
            "subscriptions for them to be considered duplicates.",
        )
        parser.add_argument(
            "--fields",
            action="store",
            dest="fields",
            type=str,
            nargs="+",
            default=self.DEFAULT_FIELDS,
            help="The fields to compare to determine if the subscriptions are"
            " identical or not.",
        )
        parser.add_argument(
            "--fix",
            action="store_true",
            default=False,
            help=(
                "Actually remove the duplicates instead of only listing" "duplicates"
            ),
        )
        parser.add_argument(
            "--scheduler-url",
            type=url_validator,
            default=settings.SCHEDULER_URL,
            help="The scheduler url to use, defaults to settings.SCHEDULER_URL",
        )
        parser.add_argument(
            "--scheduler-token",
            type=str,
            default=settings.SCHEDULER_API_TOKEN,
            help=(
                "The schedule API token, defaults to " "settings.SCHEDULER_API_TOKEN"
            ),
        )

    def second_diff(self, d1, d2):
        """
        Returns the difference, in seconds, between d1 and d2. The amount
        returned is always positive.
        """
        return abs((d1 - d2).total_seconds())

    def is_within_limits(self, limit, date, dates):
        """
        Returns True if the difference between date and any value in dates
        is less than or equal to limit.
        """
        return any((self.second_diff(date, d) <= limit for d in dates))

    def handle(self, *args, **options):
        time_delta = options["time_delta"]
        fields = options["fields"]
        fix = options["fix"]
        scheduler_token = options["scheduler_token"]
        scheduler_url = options["scheduler_url"]
        removed = 0

        scheduler_client = SchedulerApiClient(scheduler_token, scheduler_url)

        uniques = Subscription.objects.distinct(*fields)
        for unique in uniques:
            subscriptions = Subscription.objects.filter(
                **{field: getattr(unique, field) for field in fields}
            )
            subscriptions = subscriptions.filter(active=True, completed=False)
            subscriptions = subscriptions.order_by("created_at")
            if len(subscriptions) <= 1:
                continue
            dates = [subscriptions[0].created_at]
            for sub in subscriptions[1:]:
                if self.is_within_limits(time_delta, sub.created_at, dates):
                    if fix:
                        scheduler_id = sub.get_scheduler_id()
                        # NOTE: grabbing the signature of the sub before
                        #       deleting otherwise the warning below just
                        #       shows `None` as Django deletes the PK
                        #       after delete() is called
                        sub_str = str(sub)
                        sub.delete()
                        if scheduler_id:
                            scheduler_client.delete_schedule(scheduler_id)
                        else:
                            self.warning(
                                "Subscription %s has no scheduler_id." % (sub_str,)
                            )
                    else:
                        self.warning(
                            "Not removing %s, use --fix to actually remove." % (sub,)
                        )

                    removed += 1
                dates.append(sub.created_at)

        self.success("Removed %d duplicate subscriptions." % (removed,))

    def log(self, level, msg):
        self.stdout.write(level(msg))

    def warning(self, msg):
        self.log(self.style.WARNING, msg)

    def success(self, msg):
        self.log(self.style.SUCCESS, msg)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-09-12 09:23
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("contentstore", "0004_auto_20160810_0916")]

    operations = [
        migrations.AlterUniqueTogether(
            name="message",
            unique_together=set([("messageset", "sequence_number", "lang")]),
        )
    ]



from __future__ import absolute_import

import os

import celery
import raven
from django.conf import settings
from raven.contrib.celery import register_logger_signal, register_signal

# set the default Django settings module for the 'celery' program.
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "seed_stage_based_messaging.settings")


class Celery(celery.Celery):
    def on_configure(self):
        client = raven.Client(settings.STAGE_BASED_MESSAGING_SENTRY_DSN)

        # register a custom filter to filter out duplicate logs
        register_logger_signal(client)

        # hook into the Celery error handler
        register_signal(client)


app = Celery(__name__)

# Using a string here means the worker will not have to
# pickle the object when using Windows.
app.config_from_object("django.conf:settings", namespace="CELERY")
app.autodiscover_tasks()


@app.task(bind=True)
def debug_task(self):
    print("Request: {0!r}".format(self.request))



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2018-02-01 09:03
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("contentstore", "0007_auto_20171102_0950"),
        ("subscriptions", "0007_auto_20171102_0950"),
    ]

    operations = [
        migrations.CreateModel(
            name="ResendRequest",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("received_at", models.DateTimeField(auto_now_add=True)),
                ("outbound", models.UUIDField(null=True)),
                (
                    "message",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="resend_requests",
                        to="contentstore.Message",
                    ),
                ),
                (
                    "subscription",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="subscriptions.Subscription",
                    ),
                ),
            ],
        )
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-11-02 09:50
from __future__ import unicode_literals

from django.db import migrations, models

import contentstore.models


class Migration(migrations.Migration):

    dependencies = [("contentstore", "0006_messageset_channel")]

    operations = [
        migrations.AlterField(
            model_name="message",
            name="text_content",
            field=models.TextField(
                blank=True,
                null=True,
                validators=[contentstore.models.validate_special_characters],
            ),
        )
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-10-18 07:29
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("contentstore", "0008_schedule_scheduler_schedule_id")]

    operations = [
        migrations.AddField(
            model_name="messageset",
            name="label",
            field=models.CharField(
                blank=True,
                default=b"",
                max_length=100,
                verbose_name="User-readable name",
            ),
        )
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-03-22 15:34
from __future__ import unicode_literals

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("contentstore", "0002_messageset_content_type"),
        ("subscriptions", "0002_auto_20160224_0822"),
    ]

    operations = [
        migrations.RemoveField(model_name="subscription", name="messageset_id"),
        migrations.AddField(
            model_name="subscription",
            name="messageset",
            field=models.ForeignKey(
                default=1,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="subscriptions",
                to="contentstore.MessageSet",
            ),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="subscription",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="subscriptions_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="subscription",
            name="schedule",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="subscriptions",
                to="contentstore.Schedule",
            ),
        ),
        migrations.AlterField(
            model_name="subscription",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="subscriptions_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



from django import forms
from django.contrib import admin, messages
from django.shortcuts import render

from .models import BinaryContent, Message, MessageSet, Schedule


class CloneMessageSetForm(forms.Form):
    short_name = forms.CharField(label="New short name", required=True)

    def clean_short_name(self):
        short_name = self.cleaned_data["short_name"]
        if MessageSet.objects.filter(short_name=short_name).exists():
            raise forms.ValidationError("A message set already exists with this name.")
        return short_name


class MessageSetAdmin(admin.ModelAdmin):

    actions = ["clone_messageset"]

    def clone_messageset(self, request, queryset):
        if queryset.count() > 1:
            self.message_user(
                request,
                "Only 1 message set can be cloned at a time.",
                level=messages.WARNING,
            )
            return

        original = queryset.first()
        if "do_action" in request.POST:
            form = CloneMessageSetForm(request.POST)
            if form.is_valid():
                new_short_name = form.cleaned_data["short_name"]

                # NOTE: setting the PK to `None` results in a new record being
                # #     created
                clone = queryset.first()
                clone.pk = None
                clone.short_name = new_short_name
                clone.save()

                for cloned_message in original.messages.all():
                    cloned_message.pk = None
                    cloned_message.messageset = clone
                    cloned_message.save()

                self.message_user(request, "Cloned the message set")
                return
        else:
            form = CloneMessageSetForm()

        return render(
            request,
            "contentstore/messageset/clone_messageset.html",
            {
                "title": u"Clone message set: %s" % (original,),
                "object": original,
                "form": form,
            },
        )

    clone_messageset.short_description = "Clone selected message set"


admin.site.register(Schedule)
admin.site.register(MessageSet, MessageSetAdmin)
admin.site.register(Message)
admin.site.register(BinaryContent)



# Generated by Django 2.1.2 on 2018-11-26 11:04

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("contentstore", "0009_add_messageset_label")]

    operations = [
        migrations.AlterField(
            model_name="messageset",
            name="default_schedule",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="message_sets",
                to="contentstore.Schedule",
            ),
        ),
        migrations.AlterField(
            model_name="messageset",
            name="label",
            field=models.CharField(
                blank=True,
                default="",
                max_length=100,
                verbose_name="User-readable name",
            ),
        ),
        migrations.AlterField(
            model_name="messageset",
            name="next_set",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="contentstore.MessageSet",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-02-24 08:22
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("subscriptions", "0001_initial")]

    operations = [
        migrations.RenameField(
            model_name="subscription", old_name="contact", new_name="identity"
        )
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-02-11 11:04
from __future__ import unicode_literals

import uuid

import django.contrib.postgres.fields.jsonb
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="Subscription",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("contact", models.CharField(max_length=36)),
                ("version", models.IntegerField(default=1)),
                ("messageset_id", models.IntegerField()),
                ("next_sequence_number", models.IntegerField(default=1)),
                ("lang", models.CharField(max_length=6)),
                ("active", models.BooleanField(default=True)),
                ("completed", models.BooleanField(default=False)),
                ("schedule", models.IntegerField(default=1)),
                ("process_status", models.IntegerField(default=0)),
                (
                    "metadata",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        blank=True, null=True
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "created_by",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="registrations_created",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "updated_by",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="registrations_updated",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        )
    ]



#!/usr/bin/env python
import os
import sys

if __name__ == "__main__":
    os.environ.setdefault(
        "DJANGO_SETTINGS_MODULE", "seed_stage_based_messaging.settings"
    )

    from django.core.management import execute_from_command_line

    execute_from_command_line(sys.argv)



try:
    from urlparse import urlunparse
except ImportError:
    from urllib.parse import urlunparse

from datetime import timedelta
from functools import partial

from celery.exceptions import SoftTimeLimitExceeded
from celery.task import Task
from celery.utils.log import get_task_logger
from demands import HTTPServiceError
from django.conf import settings
from django.contrib.sites.shortcuts import get_current_site
from django.core import serializers
from django.core.cache import caches
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.db.models import Count, F, Q
from django.utils.timezone import now
from requests.exceptions import ConnectionError, HTTPError, Timeout
from seed_services_client import MessageSenderApiClient, SchedulerApiClient
from seed_services_client.metrics import MetricsApiClient

from contentstore.models import Message, Schedule
from seed_stage_based_messaging import utils
from seed_stage_based_messaging.celery import app

from .models import (
    BehindSubscription,
    EstimatedSend,
    ResendRequest,
    Subscription,
    SubscriptionSendFailure,
)

logger = get_task_logger(__name__)

locmem_cache = caches["locmem"]
redis_cache = caches["redis"]


def get_metric_client(session=None):
    return MetricsApiClient(
        url=settings.METRICS_URL, auth=settings.METRICS_AUTH, session=session
    )


def make_absolute_url(path):
    # NOTE: We're using the default site as set by
    #       settings.SITE_ID and the Sites framework
    site = get_current_site(None)
    return urlunparse(
        ("https" if settings.USE_SSL else "http", site.domain, path, "", "", "")
    )


class FireMetric(Task):

    """ Fires a metric using the MetricsApiClient
    """

    name = "subscriptions.tasks.fire_metric"

    def run(self, metric_name, metric_value, session=None, **kwargs):
        metric_value = float(metric_value)
        metric = {metric_name: metric_value}
        metric_client = get_metric_client(session=session)
        metric_client.fire_metrics(**metric)
        return "Fired metric <%s> with value <%s>" % (metric_name, metric_value)


fire_metric = FireMetric()


class StoreResendRequest(Task):

    """
    Task to save resend request and trigger send last message to the user.
    """

    name = "subscriptions.tasks.store_resend_request"

    def run(self, subscription_id, **kwargs):
        resend_request = ResendRequest.objects.create(subscription_id=subscription_id)

        send_current_message.delay(subscription_id, resend_request.id)

        return "Message queued for resend, subscriber: {}".format(subscription_id)


store_resend_request = StoreResendRequest()


class BaseSendMessage(Task):

    """
    Base Task for sending messages
    """

    class FailedEventRequest(Exception):

        """
        The attempted task failed because of a non-200 HTTP return
        code.
        """

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        # This function only gets called once all retries have failed, not with
        # each retry, this was tested in real life and is tested on a Celery
        # level but doesn't work with unit tests when CELERY_ALWAYS_EAGER is
        # True.
        # The function is called directly in the unit tests to make sure this
        # functionality works
        if isinstance(args[0], dict):
            subscription_id = args[0]["subscription_id"]
        else:
            subscription_id = args[0]

        SubscriptionSendFailure.objects.create(
            subscription_id=subscription_id,
            initiated_at=self.request.eta or now(),
            reason=str(exc),
            task_id=task_id,
        )
        super(BaseSendMessage, self).on_failure(exc, task_id, args, kwargs, einfo)


@app.task(
    retry_backoff=True,
    retry_jitter=True,
    max_retries=15,
    acks_late=True,
    soft_time_limit=10,
    time_limit=15,
    base=BaseSendMessage,
    bind=True,
)
def pre_send_process(self, subscription_id, resend_id=None):
    logger.debug("Locking subscription")
    key = "subscription_lock:{}".format(subscription_id)
    locked = not redis_cache.add(
        key,
        now() + timedelta(seconds=settings.SUBSCRIPTION_LOCK_TIMEOUT),
        timeout=settings.SUBSCRIPTION_LOCK_TIMEOUT,
    )
    if locked:
        retry_timestamp = redis_cache.get(key)
        logger.debug("Subscription locked, retrying at {}".format(retry_timestamp))
        self.retry(eta=retry_timestamp)

    context = {}
    if resend_id:
        context["resend_id"] = resend_id

    logger.info("Loading Subscription")
    subscription = Subscription.objects.select_related("messageset").get(
        id=subscription_id
    )

    if not subscription.is_ready_for_processing:
        if subscription.process_status == 2 or subscription.completed is True:
            # Subscription is complete
            logger.info("Subscription has completed")
            context["error"] = "Subscription has completed"

        else:
            logger.info("Message sending aborted - broken or inactive")
            # TODO: be more specific about why it aborted
            context["error"] = (
                "Message sending aborted, status <%s>" % subscription.process_status
            )
        return context

    context.update(
        {
            "subscription": serializers.serialize("json", [subscription]),
            "messageset": serializers.serialize("json", [subscription.messageset]),
        }
    )

    try:
        logger.info("Loading Message")
        next_sequence_number = subscription.next_sequence_number
        if next_sequence_number > 1 and resend_id:
            next_sequence_number -= 1

        message = locmem_cache.get_or_set(
            "message:{}:{}:{}".format(
                subscription.messageset_id, next_sequence_number, subscription.lang
            ),
            partial(
                Message.objects.get,
                messageset=subscription.messageset,
                sequence_number=next_sequence_number,
                lang=subscription.lang,
            ),
        )

        context["message"] = serializers.serialize("json", [message])
    except ObjectDoesNotExist:
        error = (
            "Missing Message: MessageSet: <%s>, Sequence Number: <%s>" ", Lang: <%s>"
        ) % (
            subscription.messageset,
            subscription.next_sequence_number,
            subscription.lang,
        )
        logger.error(error, exc_info=True)
        context["error"] = "Message sending aborted, missing message"
        return context

    return context


@app.task(
    autoretry_for=(
        HTTPError,
        ConnectionError,
        Timeout,
        HTTPServiceError,
        SoftTimeLimitExceeded,
    ),
    retry_backoff=True,
    retry_jitter=True,
    max_retries=15,
    acks_late=True,
    soft_time_limit=10,
    time_limit=15,
    base=BaseSendMessage,
)
def get_identity_address(context):
    if "error" in context:
        return context

    [deserialized_subscription] = serializers.deserialize(
        "json", context["subscription"]
    )
    subscription = deserialized_subscription.object

    to_addr = utils.get_identity_address(
        subscription.identity, use_communicate_through=True
    )

    if to_addr is None:
        logger.info("No valid recipient to_addr found")
        subscription.process_status = -1
        deserialized_subscription.save(update_fields=("process_status",))

        context["error"] = "Valid recipient could not be found"
    else:
        context["to_addr"] = to_addr

    return context


@app.task(
    autoretry_for=(
        HTTPError,
        ConnectionError,
        Timeout,
        HTTPServiceError,
        SoftTimeLimitExceeded,
    ),
    retry_backoff=True,
    retry_jitter=True,
    max_retries=15,
    acks_late=True,
    soft_time_limit=10,
    time_limit=15,
    base=BaseSendMessage,
)
def send_message(context):
    if "error" in context:
        return context

    [deserialized_subscription] = serializers.deserialize(
        "json", context["subscription"]
    )
    subscription = deserialized_subscription.object
    [messageset] = serializers.deserialize("json", context["messageset"])
    messageset = messageset.object
    [message] = serializers.deserialize("json", context["message"])
    message = message.object

    payload = {
        "to_addr": context["to_addr"],
        "to_identity": subscription.identity,
        "delivered": "false",
        "resend": "true" if "resend_id" in context else "false",
        "metadata": message.metadata,
    }

    if messageset.channel:
        payload["channel"] = messageset.channel

    prepend_next = (subscription.metadata or {}).get("prepend_next_delivery", None)
    if messageset.content_type == "text":
        logger.debug("Determining payload content")
        if prepend_next:
            logger.debug("Prepending next delivery")
            payload["content"] = "%s\n%s" % (prepend_next, message.text_content)
        else:
            logger.debug("Loading default content")
            payload["content"] = message.text_content

        if message.binary_content:
            payload["metadata"]["image_url"] = make_absolute_url(
                message.binary_content.content.url
            )

        logger.debug("text content loaded")
    else:
        if prepend_next:
            payload["metadata"]["voice_speech_url"] = [
                prepend_next,
                make_absolute_url(message.binary_content.content.url),
            ]
        else:
            payload["metadata"]["voice_speech_url"] = [
                make_absolute_url(message.binary_content.content.url)
            ]

    if messageset.id in settings.DRY_RUN_MESSAGESETS:
        logger.info("Skipping sending of message")
    else:
        logger.info("Sending message to Message Sender")
        message_sender_client = MessageSenderApiClient(
            settings.MESSAGE_SENDER_TOKEN,
            settings.MESSAGE_SENDER_URL,
            retries=5,
            timeout=settings.DEFAULT_REQUEST_TIMEOUT,
        )
        result = message_sender_client.create_outbound(payload)
        context["outbound_id"] = result["id"]

    if prepend_next:
        logger.debug("Clearing prepended message")
        subscription.metadata["prepend_next_delivery"] = None
        deserialized_subscription.save(update_fields=("metadata",))
        context["subscription"] = serializers.serialize("json", [subscription])

    logger.debug("Message queued for send. ID: <%s>" % str(context.get("outbound_id")))
    return context


@app.task(
    retry_backoff=True,
    retry_jitter=True,
    max_retries=15,
    acks_late=True,
    soft_time_limit=10,
    time_limit=15,
    base=BaseSendMessage,
)
def post_send_process(context):
    """
    Task to ensure subscription is bumped or converted
    """
    if "error" in context:
        return context

    [deserialized_subscription] = serializers.deserialize(
        "json", context["subscription"]
    )
    subscription = deserialized_subscription.object
    [messageset] = serializers.deserialize("json", context["messageset"])
    messageset = messageset.object

    # Get set max
    set_max = locmem_cache.get_or_set(
        "messageset_size:{}:{}".format(messageset.id, subscription.lang),
        messageset.messages.filter(lang=subscription.lang).count,
    )
    logger.debug("set_max calculated - %s" % set_max)

    # Compare user position to max
    if subscription.next_sequence_number == set_max:
        with transaction.atomic():
            # Mark current as completed
            logger.debug("marking current subscription as complete")
            subscription.completed = True
            subscription.active = False
            subscription.process_status = 2  # Completed
            subscription.updated_at = now()
            deserialized_subscription.save(
                update_fields=("completed", "active", "process_status", "updated_at")
            )
            # If next set defined create new subscription
            if messageset.next_set:
                logger.info("Creating new subscription for next set")
                newsub = Subscription.objects.create(
                    identity=subscription.identity,
                    lang=subscription.lang,
                    messageset=messageset.next_set,
                    schedule=messageset.next_set.default_schedule,
                )
                logger.debug("Created Subscription <%s>" % newsub.id)
    else:
        # More in this set so increment by one
        logger.debug("incrementing next_sequence_number")
        subscription.next_sequence_number = F("next_sequence_number") + 1
        subscription.updated_at = now()
        logger.debug("saving subscription")
        deserialized_subscription.save(
            update_fields=("next_sequence_number", "updated_at")
        )
    logger.debug("unlocking the subscription")
    redis_cache.delete("subscription_lock:{}".format(subscription.id))
    # return response
    return "Subscription for %s updated" % str(subscription.id)


@app.task(
    retry_backoff=True,
    retry_jitter=True,
    max_retries=15,
    acks_late=True,
    soft_time_limit=10,
    time_limit=15,
    base=BaseSendMessage,
)
def post_send_process_resend(context):
    [message] = serializers.deserialize("json", context["message"])
    message = message.object
    [deserialized_subscription] = serializers.deserialize(
        "json", context["subscription"]
    )
    subscription = deserialized_subscription.object
    resend_request = ResendRequest.objects.get(id=context["resend_id"])

    with transaction.atomic():
        if "outbound_id" in context:
            resend_request.outbound = context["outbound_id"]
        resend_request.message_id = message.id
        resend_request.save(update_fields=("outbound", "message_id"))
        subscription.process_status = 0
        subscription.updated_at = now()
        deserialized_subscription.save(update_fields=("process_status", "updated_at"))


send_next_message = (
    pre_send_process.s()
    | get_identity_address.s()
    | send_message.s()
    | post_send_process.s()
)

send_current_message = (
    pre_send_process.s()
    | get_identity_address.s()
    | send_message.s()
    | post_send_process_resend.s()
)


class ScheduleDisable(Task):

    """ Task to disable a subscription's schedule
    """

    name = "subscriptions.tasks.schedule_disable"

    def scheduler_client(self):
        return SchedulerApiClient(settings.SCHEDULER_API_TOKEN, settings.SCHEDULER_URL)

    def run(self, subscription_id, **kwargs):
        log = self.get_logger(**kwargs)
        log.info("Disabling schedule for <%s>" % (subscription_id,))
        try:
            subscription = Subscription.objects.get(id=subscription_id)
            try:
                schedule_id = subscription.metadata["scheduler_schedule_id"]
                scheduler = self.scheduler_client()
                scheduler.update_schedule(
                    subscription.metadata["scheduler_schedule_id"], {"enabled": False}
                )
                log.info(
                    "Disabled schedule <%s> on scheduler for sub <%s>"
                    % (schedule_id, subscription_id)
                )
                return True
            except Exception:
                log.info("Schedule id not saved in subscription metadata")
                return False
        except ObjectDoesNotExist:
            logger.error("Missing Subscription", exc_info=True)
        except SoftTimeLimitExceeded:
            logger.error(
                "Soft time limit exceed processing schedule create " "via Celery.",
                exc_info=True,
            )
        return False


schedule_disable = ScheduleDisable()


class ScheduledMetrics(Task):

    """ Fires off tasks for all the metrics that should run
        on a schedule
    """

    name = "subscriptions.tasks.scheduled_metrics"

    def run(self, **kwargs):
        globs = globals()  # execute globals() outside for loop for efficiency
        for metric in settings.METRICS_SCHEDULED_TASKS:
            globs[metric].apply_async()

        return "%d Scheduled metrics launched" % len(settings.METRICS_SCHEDULED_TASKS)


scheduled_metrics = ScheduledMetrics()


class FireWeekEstimateLast(Task):
    """Fires week estimated send counts.
    """

    name = "subscriptions.tasks.fire_week_estimate_last"

    def run(self):
        schedules = Schedule.objects.filter(
            subscriptions__active=True,
            subscriptions__completed=False,
            subscriptions__process_status=0,
        ).annotate(total_subs=Count("subscriptions"))
        totals = {0: 0, 1: 0, 2: 0, 3: 0, 4: 0, 5: 0, 6: 0}
        for schedule in schedules:
            for day in range(7):
                if str(day) in schedule.day_of_week or "*" in schedule.day_of_week:
                    totals[day] = totals[day] + schedule.total_subs

        # Django's datetime's weekday method has Monday = 0
        # whereas the cron format used in the schedules has Sunday = 0
        sunday = totals.pop(0)
        totals[7] = sunday
        totals = {(k - 1): v for k, v in totals.items()}

        today = now()
        for dow, total in totals.items():
            # Only fire the metric for today or days in the future so that
            # estimates for the week don't get updated after the day in
            # question.
            if dow >= (today.weekday()):
                fire_metric.apply_async(
                    kwargs={
                        "metric_name": "subscriptions.send.estimate.%s.last" % dow,
                        "metric_value": total,
                    }
                )


fire_week_estimate_last = FireWeekEstimateLast()


class FireDailySendEstimate(Task):
    """Fires daily estimated send counts.
    """

    name = "subscriptions.tasks.fire_daily_send_estimate"

    def run(self):
        # Django's datetime's weekday method has Monday = 0
        # whereas the cron format used in the schedules has Sunday = 0
        day = now().weekday() + 1

        schedules = (
            Schedule.objects.filter(
                Q(day_of_week__contains=day) | Q(day_of_week__contains="*"),
                subscriptions__active=True,
                subscriptions__completed=False,
                subscriptions__process_status=0,
            )
            .values("subscriptions__messageset")
            .annotate(
                total_subs=Count("subscriptions"),
                total_unique=Count("subscriptions__identity", distinct=True),
            )
        )

        for schedule in schedules:
            EstimatedSend.objects.get_or_create(
                send_date=now().date(),
                messageset_id=schedule["subscriptions__messageset"],
                estimate_subscriptions=schedule["total_subs"],
                estimate_identities=schedule["total_unique"],
            )


fire_daily_send_estimate = FireDailySendEstimate()


class RequeueFailedTasks(Task):

    """
    Task to requeue failed schedules.
    """

    name = "subscriptions.tasks.requeue_failed_tasks"

    def run(self, **kwargs):
        log = self.get_logger(**kwargs)
        failures = SubscriptionSendFailure.objects
        log.info(
            "Attempting to requeue <%s> failed Subscription sends"
            % failures.all().count()
        )
        for failure in failures.iterator():
            subscription_id = str(failure.subscription_id)
            # Cleanup the failure before requeueing it.
            failure.delete()
            send_next_message.delay(subscription_id)


requeue_failed_tasks = RequeueFailedTasks()


@app.task
def calculate_subscription_lifecycle(subscription_id):
    """
    Calculates the expected lifecycle position the subscription in
    subscription_ids, and creates a BehindSubscription entry for them.

    Args:
        subscription_id (str): ID of subscription to calculate lifecycle for
    """
    subscription = Subscription.objects.select_related("messageset", "schedule").get(
        id=subscription_id
    )
    behind = subscription.messages_behind()
    if behind == 0:
        return

    current_messageset = subscription.messageset
    current_sequence_number = subscription.next_sequence_number
    end_subscription = Subscription.fast_forward_lifecycle(subscription, save=False)[-1]
    BehindSubscription.objects.create(
        subscription=subscription,
        messages_behind=behind,
        current_messageset=current_messageset,
        current_sequence_number=current_sequence_number,
        expected_messageset=end_subscription.messageset,
        expected_sequence_number=end_subscription.next_sequence_number,
    )


@app.task
def find_behind_subscriptions():
    """
    Finds any subscriptions that are behind according to where they should be,
    and creates a BehindSubscription entry for them.
    """
    subscriptions = Subscription.objects.filter(
        active=True, completed=False, process_status=0
    ).values_list("id", flat=True)
    for subscription_id in subscriptions.iterator():
        calculate_subscription_lifecycle.delay(str(subscription_id))



"""
Tests for the contentstore tasks
"""

from unittest.mock import patch

from django.test import TestCase

from contentstore.models import MessageSet, Schedule
from contentstore.tasks import queue_subscription_send
from seed_stage_based_messaging import test_utils as utils
from subscriptions.models import Subscription


class QueueSubscriptionSendTaskTests(TestCase):
    """
    Tests for the queue subscription send task
    """

    def setUp(self):
        utils.disable_signals()

    def tearDown(self):
        utils.enable_signals()

    @patch("contentstore.tasks.send_next_message")
    def test_queue_subscription_send(self, send_next_message):
        """
        The queue subscription send task should run the send next message task
        for each active and valid subscription for the specified schedule.
        """
        schedule1 = Schedule.objects.create()
        schedule2 = Schedule.objects.create()
        messageset = MessageSet.objects.create(default_schedule=schedule2)

        # Subscriptions that shouldn't be run
        Subscription.objects.create(messageset=messageset, schedule=schedule2)
        Subscription.objects.create(
            messageset=messageset, schedule=schedule1, active=False
        )
        Subscription.objects.create(
            messageset=messageset, schedule=schedule1, completed=True
        )
        Subscription.objects.create(
            messageset=messageset, schedule=schedule1, process_status=1
        )

        # Subscriptions that should be run
        subscription = Subscription.objects.create(
            messageset=messageset, schedule=schedule1
        )

        queue_subscription_send(str(schedule1.id))
        send_next_message.delay.assert_called_once_with(str(subscription.id))



# -*- coding: utf-8 -*-

import os.path
import re
from datetime import datetime

from croniter import croniter
from django.conf import settings
from django.contrib.postgres.fields import JSONField
from django.db import models
from django.shortcuts import reverse
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _
from rest_framework.serializers import ValidationError

try:
    from urlparse import urljoin
except ImportError:
    from urllib.parse import urljoin


def validate_special_characters(value):
    unnecessary_special_characters = [u"“", u"”", u"’", u"–"]
    errors = []

    for character in value:
        if character in unnecessary_special_characters:
            errors.append(character)

    if errors:
        raise ValidationError(
            u'Text "{0}" has special characters which can be replaced: {1}'.format(
                value, " ".join(errors)
            )
        )


@python_2_unicode_compatible
class Schedule(models.Model):

    """
    Schedules (sometimes referred to as Protocols) are the method used to
    define the rate and frequency at which the messages are sent to
    the recipient
    """

    minute = models.CharField(_("minute"), max_length=64, default="*")
    hour = models.CharField(_("hour"), max_length=64, default="*")
    day_of_week = models.CharField(_("day of week"), max_length=64, default="*")
    day_of_month = models.CharField(_("day of month"), max_length=64, default="*")
    month_of_year = models.CharField(_("month of year"), max_length=64, default="*")
    scheduler_schedule_id = models.UUIDField(
        _("scheduler schedule id"), null=True, blank=True, default=None
    )

    class Meta:
        verbose_name = _("schedule")
        verbose_name_plural = _("schedules")
        ordering = ["month_of_year", "day_of_month", "day_of_week", "hour", "minute"]

    @property
    def cron_string(self):
        return "{minute} {hour} {dom} {moy} {dow}".format(
            minute=self.rfield(self.minute),
            hour=self.rfield(self.hour),
            dom=self.rfield(self.day_of_month),
            moy=self.rfield(self.month_of_year),
            dow=self.rfield(self.day_of_week),
        )

    def rfield(self, s):
        return s and str(s).replace(" ", "") or "*"

    def __str__(self):
        return "{0} {1} {2} {3} {4} (m/h/d/dM/MY)".format(
            self.rfield(self.minute),
            self.rfield(self.hour),
            self.rfield(self.day_of_week),
            self.rfield(self.day_of_month),
            self.rfield(self.month_of_year),
        )

    @property
    def send_url(self):
        """
        Returns the URL to send this schedule's subscriptions
        """
        callback_path = reverse("schedule-detail", args=[str(self.id)])
        callback_path = "{}send/".format(callback_path)
        return urljoin(settings.STAGE_BASED_MESSAGING_URL, callback_path)

    @property
    def scheduler_format(self):
        """
        Returns the format that the scheduler service is expecting.
        """
        return {
            "frequency": None,
            "cron_definition": self.cron_string,
            "endpoint": self.send_url,
            "auth_token": settings.SCHEDULER_INBOUND_API_TOKEN,
        }

    def get_run_times_between(self, start, end):
        """Gets a list of datetimes for when this cron schedule would be
        run between the given start and end datetimes.
        """
        dates = []
        for dt in croniter(self.cron_string, start, ret_type=datetime):
            if dt > end:
                break
            dates.append(dt)
        return dates


@python_2_unicode_compatible
class MessageSet(models.Model):

    """
        Details about a set of messages that a recipient can be sent on
        a particular schedule
    """

    CONTENT_TYPES = (("text", "Text"), ("audio", "Audio"))

    short_name = models.CharField(_("Short name"), max_length=100, unique=True)
    label = models.CharField(
        _("User-readable name"), max_length=100, blank=True, default=""
    )
    notes = models.TextField(_("Notes"), null=True, blank=True)
    channel = models.CharField(_("Channel"), max_length=64, null=True, blank=True)
    next_set = models.ForeignKey(
        "self", null=True, blank=True, on_delete=models.SET_NULL
    )
    default_schedule = models.ForeignKey(
        Schedule, related_name="message_sets", null=False, on_delete=models.PROTECT
    )
    content_type = models.CharField(
        choices=CONTENT_TYPES, max_length=20, default="text"
    )
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __str__(self):
        return "%s" % self.short_name

    def get_messageset_max(self, lang):
        return self.messages.filter(lang=lang).count()

    def get_all_run_dates(self, start, lang, schedule=None, initial=None):
        """Returns the complete list of dates this MessageSet would run on given
        a start datetime, language, Schedule and initial message. If no
        Schedule is passed it will use the configured default_schedule. If no
        Initial is passed it assumes the full set will run.
        """
        if schedule is None:
            schedule = self.default_schedule
        dates = []
        set_max = self.get_messageset_max(lang)
        iters = initial if initial else 1  # start iterator at initial message
        for dt in croniter(schedule.cron_string, start, ret_type=datetime):
            if iters > set_max:
                break
            iters = iters + 1
            dates.append(dt)
        return dates


def generate_new_filename(instance, filename):
    ext = os.path.splitext(filename)[-1]  # get file extension
    return "%s%s" % (datetime.now().strftime("%Y%m%d%H%M%S%f"), ext)


@python_2_unicode_compatible
class BinaryContent(models.Model):
    """
        File store for reference in messages. Storage method handle by
        settings file.
    """

    content = models.FileField(upload_to=generate_new_filename, max_length=100)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __str__(self):
        return "%s" % (self.content.path.split("/")[-1])


@python_2_unicode_compatible
class Message(models.Model):

    """
        Messages that a recipient can be sent
    """

    messageset = models.ForeignKey(
        MessageSet, related_name="messages", null=False, on_delete=models.CASCADE
    )
    sequence_number = models.IntegerField(null=False, blank=False)
    lang = models.CharField(max_length=6, null=False, blank=False)
    metadata = JSONField(default=dict)
    text_content = models.TextField(
        null=True, blank=True, validators=[validate_special_characters]
    )
    binary_content = models.ForeignKey(
        BinaryContent,
        related_name="message",
        null=True,
        blank=True,
        on_delete=models.CASCADE,
    )
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    WHATSAPP_INVALID_VALUE_PATTERN = re.compile(r"(\n|\t|\s{4})")

    class Meta:
        ordering = ["sequence_number"]
        unique_together = ("messageset", "sequence_number", "lang")

    def clean(self):
        # Don't allow messages to have neither a text or binary content
        if any([self.text_content, self.binary_content]) is False:
            raise ValidationError(_("Messages must have text or file attached"))
        # Check for WhatsApp disallowed characters
        if "whatsapp" in (
            self.messageset.channel or ""
        ).lower() and self.WHATSAPP_INVALID_VALUE_PATTERN.findall(self.text_content):
            raise ValidationError(
                _(
                    "Text for WhatsApp must not include newline, tabs, or 4 "
                    "consecutive spaces."
                )
            )

    def save(self, *args, **kwargs):
        self.clean()
        super(Message, self).save(*args, **kwargs)

    def __str__(self):
        return _("Message %s in %s from %s") % (
            self.sequence_number,
            self.lang,
            self.messageset.short_name,
        )



import os

import paramiko
import requests
from celery.task import Task
from django.conf import settings
from django.db.models.signals import post_save
from django.utils._os import abspathu
from seed_services_client.scheduler import SchedulerApiClient
from sftpclone import sftpclone

from contentstore.models import Schedule
from contentstore.signals import schedule_saved
from subscriptions.models import Subscription
from subscriptions.tasks import make_absolute_url, send_next_message

from .models import BinaryContent


class SyncAudioMessages(Task):
    def _get_existing_files(self, root, host, username, password, port):
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(host, username=username, password=password, port=port)
        ftp = ssh.open_sftp()
        existing_files = ftp.listdir(root)
        return existing_files

    def run(self, **kwargs):
        if settings.AUDIO_FTP_HOST:
            src = abspathu(settings.MEDIA_ROOT)
            if not src.endswith("/"):
                src += "/"

            root = settings.AUDIO_FTP_ROOT
            host = settings.AUDIO_FTP_HOST
            username = settings.AUDIO_FTP_USER
            password = settings.AUDIO_FTP_PASS
            port = int(settings.AUDIO_FTP_PORT)

            existing_files = self._get_existing_files(
                root, host, username, password, port
            )

            delete_files = []

            files = BinaryContent.objects.all()
            for item in files.iterator():

                if item.content.name not in existing_files:
                    r = requests.get(make_absolute_url(item.content.url))

                    local_path = "{}{}".format(src, item.content.name)
                    with open(local_path, "wb") as f:
                        f.write(r.content)

                    delete_files.append(local_path)

            cloner = sftpclone.SFTPClone(
                src,
                "{}:{}@{}:{}".format(username, password, host, root),
                port=port,
                delete=False,
            )

            cloner.run()

            for item in delete_files:
                os.remove(item)


sync_audio_messages = SyncAudioMessages()


class QueueSubscriptionSend(Task):
    """
    Queues the send next message task for all of the subscriptions tied to
    the schedule.
    """

    name = "contentstore.tasks.queue_subscription_send"

    def run(self, schedule_id, **kwargs):
        """
        Arguments:
            schedule_id {int} -- The schedule to send messages for
        """
        subscriptions = Subscription.objects.filter(
            schedule_id=schedule_id, active=True, completed=False, process_status=0
        ).values("id")
        for subscription in subscriptions.iterator():
            send_next_message.delay(str(subscription["id"]))


queue_subscription_send = QueueSubscriptionSend()


class SyncSchedule(Task):
    """
    Task for synchronising schedules to the scheduler service
    """

    name = "contentstore.tasks.sync_schedule"
    scheduler = SchedulerApiClient(settings.SCHEDULER_API_TOKEN, settings.SCHEDULER_URL)

    def run(self, schedule_id, **kwargs):
        """
        Synchronises the schedule specified by the ID `schedule_id` to the
        scheduler service.

        Arguments:
            schedule_id {str} -- The ID of the schedule to sync
        """
        log = self.get_logger(**kwargs)

        try:
            schedule = Schedule.objects.get(id=schedule_id)
        except Schedule.DoesNotExist:
            log.error("Missing Schedule %s", schedule_id, exc_info=True)

        if schedule.scheduler_schedule_id is None:
            # Create the new schedule
            result = self.scheduler.create_schedule(schedule.scheduler_format)
            schedule.scheduler_schedule_id = result["id"]
            # Disable update signal here to avoid calling twice
            post_save.disconnect(schedule_saved, sender=Schedule)
            schedule.save(update_fields=("scheduler_schedule_id",))
            post_save.connect(schedule_saved, sender=Schedule)

            log.info(
                "Created schedule %s on scheduler for schedule %s",
                schedule.scheduler_schedule_id,
                schedule.id,
            )
        else:
            # Update the existing schedule
            result = self.scheduler.update_schedule(
                str(schedule.scheduler_schedule_id), schedule.scheduler_format
            )
            log.info(
                "Updated schedule %s on scheduler for schedule %s",
                schedule.scheduler_schedule_id,
                schedule.id,
            )


sync_schedule = SyncSchedule()


class DeactivateSchedule(Task):
    """
    Task for deavtivating schedules in the scheduler service
    """

    name = "contentstore.tasks.deactivate_schedule"
    scheduler = SchedulerApiClient(settings.SCHEDULER_API_TOKEN, settings.SCHEDULER_URL)

    def run(self, scheduler_schedule_id, **kwargs):
        """
        Deactivates the schedule specified by the ID `scheduler_schedule_id` in
        the scheduler service.

        Arguments:
            scheduler_schedule_id {str} -- The ID of the schedule to deactivate
        """
        log = self.get_logger(**kwargs)

        self.scheduler.update_schedule(scheduler_schedule_id, {"active": False})
        log.info(
            "Deactivated schedule %s in the scheduler service", scheduler_schedule_id
        )


deactivate_schedule = DeactivateSchedule()



from demands import HTTPServiceError
from django.conf import settings
from django.core.management.base import BaseCommand
from django.core.validators import URLValidator
from requests import exceptions
from seed_services_client import HubApiClient, IdentityStoreApiClient

from subscriptions.models import Subscription


def url_validator(url_str):
    URLValidator(url_str)
    return url_str


class Command(BaseCommand):
    help = (
        "Searches all subscription with status 5 and updates the linked "
        "identity and registration object. This is for excluding the "
        "relevant items from reports"
    )

    def add_arguments(self, parser):
        parser.add_argument("--hub-url", type=url_validator, help="The HUB url to use")
        parser.add_argument("--hub-token", type=str, help=("The HUB API token"))
        parser.add_argument(
            "--identity-url",
            type=url_validator,
            default=settings.IDENTITY_STORE_URL,
            help="The Identity Store url to use",
        )
        parser.add_argument(
            "--identity-token",
            type=str,
            default=settings.IDENTITY_STORE_TOKEN,
            help=("The Identity Store API token"),
        )
        parser.add_argument(
            "--hub-identity-field",
            type=str,
            default="mother_id",
            help=("The HUB Identity field"),
        )

    def handle(self, *args, **options):
        hub_url = options["hub_url"]
        hub_token = options["hub_token"]
        id_url = options["identity_url"]
        id_token = options["identity_token"]
        hub_id_field = options["hub_identity_field"]

        if not hub_url or not hub_token:
            self.warning("hub-url and hub-token is required.")
            return

        hubApi = HubApiClient(hub_token, hub_url)
        idApi = IdentityStoreApiClient(id_token, id_url)

        subscriptions = Subscription.objects.filter(process_status=5)

        reg_count = 0
        id_count = 0
        for subscription in subscriptions.iterator():
            identity = subscription.identity

            # find and update linked registrations
            registrations = hubApi.get_registrations({hub_id_field: identity})

            for registration in registrations["results"]:

                if not registration["data"].get("exclude_report", False):
                    registration["data"]["exclude_report"] = True

                    try:
                        hubApi.update_registration(
                            registration["id"], {"data": registration["data"]}
                        )
                        reg_count += 1
                    except exceptions.ConnectionError as exc:
                        self.warning("Connection error to Hub API: {}".format(exc))
                    except HTTPServiceError as exc:
                        self.warning(
                            "Invalid Hub API response({}): {}".format(
                                exc.response.status_code, exc.response.url
                            )
                        )

            # find and update linked identities
            identity = idApi.get_identity(identity)

            if not identity["details"].get("exclude_report", False):
                identity["details"]["exclude_report"] = True

                try:
                    idApi.update_identity(
                        identity["id"], {"details": identity["details"]}
                    )
                    id_count += 1
                except exceptions.ConnectionError as exc:
                    self.warning("Connection error to Identity API: {}".format(exc))
                except HTTPServiceError as exc:
                    self.warning(
                        "Invalid Identity Store API response({}): {}".format(
                            exc.response.status_code, exc.response.url
                        )
                    )

        self.success(
            "Updated %s identities and %s registrations." % (id_count, reg_count)
        )

    def log(self, level, msg):
        self.stdout.write(level(msg))

    def warning(self, msg):
        self.log(self.style.WARNING, msg)

    def success(self, msg):
        self.log(self.style.SUCCESS, msg)



import random
import re

from django.conf import settings
from seed_services_client import IdentityStoreApiClient

from contentstore.models import MessageSet

NORMALISE_METRIC_RE = re.compile(r"\W+")

identity_store_client = IdentityStoreApiClient(
    settings.IDENTITY_STORE_TOKEN,
    settings.IDENTITY_STORE_URL,
    retries=5,
    timeout=settings.DEFAULT_REQUEST_TIMEOUT,
)


def get_identity(identity_uuid):
    return identity_store_client.get_identity(identity_uuid)


def normalise_metric_name(name):
    """
    Replaces all non-alphanumeric characters with underscores.
    """
    return NORMALISE_METRIC_RE.sub("_", name).rstrip("_").lstrip("_")


def get_identity_address(identity_uuid, use_communicate_through=False):
    params = {"default": True}
    if use_communicate_through:
        params["use_communicate_through"] = True

    return identity_store_client.get_identity_address(identity_uuid, params=params)


def get_available_metrics():
    available_metrics = []
    available_metrics.extend(settings.METRICS_REALTIME)
    available_metrics.extend(settings.METRICS_SCHEDULED)

    for messageset in MessageSet.objects.all().iterator():
        send_type = normalise_metric_name(messageset.content_type)
        ms_name = normalise_metric_name(messageset.short_name)
        available_metrics.append("message.{}.{}.sum".format(send_type, ms_name))
        available_metrics.append("message.{}.sum".format(send_type))

    return available_metrics


def calculate_retry_delay(attempt, max_delay=300):
    """Calculates an exponential backoff for retry attempts with a small
    amount of jitter."""
    delay = int(random.uniform(2, 4) ** attempt)
    if delay > max_delay:
        # After reaching the max delay, stop using expontential growth
        # and keep the delay nearby the max.
        delay = int(random.uniform(max_delay - 20, max_delay + 20))
    return delay



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-02-09 13:14
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models

import contentstore.models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="BinaryContent",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "content",
                    models.FileField(
                        upload_to=contentstore.models.generate_new_filename
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name="Message",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("sequence_number", models.IntegerField()),
                ("lang", models.CharField(max_length=6)),
                ("text_content", models.TextField(blank=True, null=True)),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "binary_content",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="message",
                        to="contentstore.BinaryContent",
                    ),
                ),
            ],
            options={"ordering": ["sequence_number"]},
        ),
        migrations.CreateModel(
            name="MessageSet",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "short_name",
                    models.CharField(
                        max_length=20, unique=True, verbose_name="Short name"
                    ),
                ),
                (
                    "notes",
                    models.TextField(blank=True, null=True, verbose_name="Notes"),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name="Schedule",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "minute",
                    models.CharField(default="*", max_length=64, verbose_name="minute"),
                ),
                (
                    "hour",
                    models.CharField(default="*", max_length=64, verbose_name="hour"),
                ),
                (
                    "day_of_week",
                    models.CharField(
                        default="*", max_length=64, verbose_name="day of week"
                    ),
                ),
                (
                    "day_of_month",
                    models.CharField(
                        default="*", max_length=64, verbose_name="day of month"
                    ),
                ),
                (
                    "month_of_year",
                    models.CharField(
                        default="*", max_length=64, verbose_name="month of year"
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "schedules",
                "ordering": [
                    "month_of_year",
                    "day_of_month",
                    "day_of_week",
                    "hour",
                    "minute",
                ],
                "verbose_name": "schedule",
            },
        ),
        migrations.AddField(
            model_name="messageset",
            name="default_schedule",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="message_sets",
                to="contentstore.Schedule",
            ),
        ),
        migrations.AddField(
            model_name="messageset",
            name="next_set",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="contentstore.MessageSet",
            ),
        ),
        migrations.AddField(
            model_name="message",
            name="messageset",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="messages",
                to="contentstore.MessageSet",
            ),
        ),
    ]



from seed_stage_based_messaging.settings import *  # noqa: F403

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "TESTSEKRET"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

TEMPLATE_DEBUG = True

CELERY_TASK_EAGER_PROPAGATES = True
CELERY_TASK_ALWAYS_EAGER = True

SCHEDULER_URL = "http://seed-scheduler/api/v1"
SCHEDULER_API_TOKEN = "REPLACEME"

IDENTITY_STORE_URL = "http://seed-identity-store/api/v1"
IDENTITY_STORE_TOKEN = "REPLACEME"

MESSAGE_SENDER_URL = "http://seed-message-sender/api/v1"
MESSAGE_SENDER_TOKEN = "REPLACEME"

METRICS_URL = "http://metrics-url"
METRICS_AUTH_TOKEN = "REPLACEME"

PASSWORD_HASHERS = ("django.contrib.auth.hashers.MD5PasswordHasher",)


# REST Framework conf defaults
REST_FRAMEWORK["PAGE_SIZE"] = 2  # noqa: F405

AUDIO_FTP_HOST = "localhost"
AUDIO_FTP_PORT = "2222"
AUDIO_FTP_USER = "test"
AUDIO_FTP_PASS = "secret"
AUDIO_FTP_ROOT = "test_directory"



# Generated by Django 2.1.2 on 2018-11-27 07:47

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("subscriptions", "0009_auto_20181126_1111")]

    operations = [
        migrations.AlterModelOptions(
            name="behindsubscription",
            options={
                "permissions": (
                    ("can_find_behind_subscriptions", "Can find behind subscriptions"),
                )
            },
        )
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-08-10 09:16
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("contentstore", "0003_auto_20160513_1115")]

    operations = [
        migrations.AlterField(
            model_name="message",
            name="binary_content",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="message",
                to="contentstore.BinaryContent",
            ),
        )
    ]



from unittest.mock import patch

from django.core.management import call_command
from django.test import TestCase
from six import StringIO

from contentstore.models import Schedule
from seed_stage_based_messaging import test_utils as utils


class SyncSchedulesTests(TestCase):
    @patch("contentstore.management.commands.sync_schedules.sync_schedule")
    def test_schedule_sync_called(self, sync_task):
        """
        The sync schedules management command should call the sync schedule
        task for every schedule.
        """
        utils.disable_signals()
        schedule = Schedule.objects.create()
        utils.enable_signals()

        out = StringIO()
        call_command("sync_schedules", stdout=out)

        sync_task.assert_called_once_with(str(schedule.id))
        self.assertIn(str(schedule.id), out.getvalue())
        self.assertIn("Synchronised 1 schedule/s", out.getvalue())



from contextlib import contextmanager
from datetime import datetime, timedelta
from unittest.mock import patch

from django.db.models import signals
from django.test import TestCase

from contentstore.models import Message, MessageSet, Schedule
from contentstore.signals import schedule_saved
from subscriptions.models import BehindSubscription, Subscription
from subscriptions.tasks import (
    calculate_subscription_lifecycle,
    find_behind_subscriptions,
    post_send_process,
    pre_send_process,
)


@contextmanager
def disable_signal(signal_name, signal_fn, sender):
    signal = getattr(signals, signal_name)
    signal.disconnect(signal_fn, sender)
    try:
        yield None
    finally:
        signal.connect(signal_fn, sender)


class TestFindBehindSubscriptionsTask(TestCase):
    def test_find_behind_subscriptions_queues_tasks(self):
        """
        find_behind_subscriptions should queue processing for all the
        subscriptions that are active
        """
        with disable_signal("post_save", schedule_saved, Schedule):
            schedule = Schedule.objects.create()
        messageset = MessageSet.objects.create(default_schedule=schedule)
        sub_valid = Subscription.objects.create(
            schedule=schedule, messageset=messageset
        )
        Subscription.objects.create(
            schedule=schedule, messageset=messageset, active=False
        )
        Subscription.objects.create(
            schedule=schedule, messageset=messageset, completed=True
        )
        Subscription.objects.create(
            schedule=schedule, messageset=messageset, process_status=-1
        )

        with patch.object(calculate_subscription_lifecycle, "delay") as p:
            find_behind_subscriptions.delay()
        p.assert_called_once_with(str(sub_valid.id))


class TestCalculateSubscriptionLifecycle(TestCase):
    def test_calculate_subscription_lifecycle_correct(self):
        """
        If the subscription is in the correct place, then no
        BehindSubscriptions should be created
        """
        with disable_signal("post_save", schedule_saved, Schedule):
            schedule = Schedule.objects.create()
        messageset = MessageSet.objects.create(default_schedule=schedule)
        for i in range(10):
            Message.objects.create(
                messageset=messageset, text_content=str(i), sequence_number=i
            )
        subscription = Subscription.objects.create(
            schedule=schedule, messageset=messageset
        )

        # Ensure that we make the minumum amount of requests needed
        # 1: Get the subscription with related schedule and messageset
        # 2: Get the count of messages for the messageset
        with self.assertNumQueries(2):
            calculate_subscription_lifecycle.delay(str(subscription.id))
        self.assertEqual(BehindSubscription.objects.count(), 0)

    def test_calculate_subscription_lifecycle_behind(self):
        """
        If the subscription is behind, then a BehindSubscription should be
        created with the correct details
        """
        with disable_signal("post_save", schedule_saved, Schedule):
            schedule = Schedule.objects.create(minute=0)
        messageset = MessageSet.objects.create(default_schedule=schedule)
        for i in range(10):
            Message.objects.create(
                messageset=messageset, text_content=str(i), sequence_number=i
            )
        subscription = Subscription.objects.create(
            schedule=schedule, messageset=messageset
        )
        subscription.created_at = datetime.now() - timedelta(hours=2)
        subscription.save()

        calculate_subscription_lifecycle.delay(str(subscription.id))
        [behind] = BehindSubscription.objects.all()
        self.assertEqual(behind.subscription, subscription)
        self.assertEqual(behind.messages_behind, 2)
        self.assertEqual(behind.current_messageset, subscription.messageset)
        self.assertEqual(
            behind.current_sequence_number, subscription.next_sequence_number
        )
        self.assertEqual(behind.expected_messageset, subscription.messageset)
        self.assertEqual(behind.expected_sequence_number, 3)


class CachedMessageLookupTests(TestCase):
    def test_caching_message_lookup_working(self):
        """
        Ensure that the second time we make a request, there's no database hit
        """
        with disable_signal("post_save", schedule_saved, Schedule):
            schedule = Schedule.objects.create(minute=0)
        messageset = MessageSet.objects.create(default_schedule=schedule)
        for i in range(10):
            Message.objects.create(
                messageset=messageset, text_content=str(i), sequence_number=i
            )
        subscription_1 = Subscription.objects.create(
            schedule=schedule, messageset=messageset
        )
        subscription_2 = Subscription.objects.create(
            schedule=schedule, messageset=messageset
        )

        with self.assertNumQueries(2):
            pre_send_process(subscription_1.id)

        with self.assertNumQueries(1):
            pre_send_process(subscription_2.id)

    def test_cache_message_count_working(self):
        """
        Ensure that the second time we do a message_count, there's no database hit
        """
        with disable_signal("post_save", schedule_saved, Schedule):
            schedule = Schedule.objects.create(minute=0)
        messageset = MessageSet.objects.create(default_schedule=schedule)
        for i in range(3):
            Message.objects.create(
                messageset=messageset, text_content=str(i), sequence_number=i
            )
        subscription = Subscription.objects.create(
            schedule=schedule, messageset=messageset
        )

        res = pre_send_process(subscription.id)

        with self.assertNumQueries(2):
            post_send_process(res)

        with self.assertNumQueries(1):
            post_send_process(res)



from django.db.models.signals import post_delete, post_save
from django.test import TestCase

from contentstore.models import Schedule
from contentstore.signals import schedule_deleted, schedule_saved

from .utils import normalise_metric_name


class NormaliseMetricNameTest(TestCase):
    def test_normalise_metric_name(self):
        """
        The normalise_metric_name function should replace all non-alphanumeric
        with underscores.
        """
        self.assertEqual(normalise_metric_name("foo^& bar"), "foo_bar")
        self.assertEqual(normalise_metric_name("foo   bar"), "foo_bar")
        self.assertEqual(normalise_metric_name("_foo!bar,"), "foo_bar")


post_save_signals = ((schedule_saved, Schedule),)


post_delete_signals = ((schedule_deleted, Schedule),)


def disable_signals():
    for (signal, model) in post_save_signals:
        post_save.disconnect(signal, sender=model)
    for (signal, model) in post_delete_signals:
        post_delete.disconnect(signal, sender=model)


def enable_signals():
    for (signal, model) in post_save_signals:
        post_save.connect(signal, sender=model)
    for (signal, model) in post_delete_signals:
        post_delete.connect(signal, sender=model)



from django.test import TestCase
from django.urls import reverse
from rest_framework import status


class InternalOnlyTests(TestCase):
    def test_internal_access(self):
        """
        Internal access should be allowed
        """
        url = reverse("metrics")
        response = self.client.get(url, HTTP_X_FORWARDED_FOR="1.2.3.4")
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_external_access(self):
        """
        External access through the load balancer should be blocked
        """
        url = reverse("metrics")
        response = self.client.get(url, HTTP_X_FORWARDED_FOR="1.2.3.4, 4.3.2.1")
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)



# Generated by Django 2.1.2 on 2019-02-20 15:47

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("contentstore", "0010_auto_20181126_1104")]

    operations = [
        migrations.AddField(
            model_name="message",
            name="metadata",
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        )
    ]



# -*- coding: utf-8 -*-

from django.test import TestCase
from rest_framework.serializers import ValidationError

from seed_stage_based_messaging import test_utils as utils

from ..models import Message, validate_special_characters
from .test_general import MessageSetTestMixin


class TestValidators(TestCase):
    def test_special_characters_allows_normal_text(self):
        self.assertEqual(validate_special_characters("ascii text!"), None)

    def test_special_characters_rejects_bad_chars(self):
        with self.assertRaises(ValidationError):
            validate_special_characters(u'I said “hello"')

        with self.assertRaises(ValidationError):
            validate_special_characters(u'I said "hello”')

        with self.assertRaises(ValidationError):
            validate_special_characters(u"It’s over there")

        with self.assertRaises(ValidationError):
            validate_special_characters(u"Hello –")


class TestMessage(MessageSetTestMixin, TestCase):
    def setUp(self):
        utils.disable_signals()

    def tearDown(self):
        utils.enable_signals()

    def test_raises_validation_error_for_special_chars(self):
        messageset = self.make_messageset()
        special_chars = u"This text has a special character – there"

        with self.assertRaises(ValidationError):
            m = Message.objects.create(
                sequence_number=1,
                messageset_id=messageset.id,
                text_content=special_chars,
            )
            m.full_clean()

    def test_raises_validation_error_for_whatsapp_invalid_chars(self):
        """
        Should raise a validation error if any of the disallowed characters
        are present in the text content.
        """
        messageset = self.make_messageset(channel="TEST_WHATSAPP_CHANNEL")
        messages = [
            "Message with \n newline",
            "Message with \t tab",
            "Message with    four spaces",
        ]
        for message in messages:
            with self.assertRaises(ValidationError):
                Message.objects.create(
                    sequence_number=1, messageset_id=messageset.id, text_content=message
                )



from django.conf.urls import include, url
from rest_framework import routers

from . import views

router = routers.DefaultRouter()
router.register(r"subscriptions", views.SubscriptionViewSet)
router.register(r"failed-tasks", views.FailedTaskViewSet)
router.register(r"behind-subscriptions", views.BehindSubscriptionViewSet)

# Wire up our API using automatic URL routing.
# Additionally, we include login URLs for the browseable API.
urlpatterns = [
    url(r"^api/v1/", include(router.urls)),
    url(
        r"^api/v1/subscriptions/(?P<subscription_id>.+)/send$",
        views.SubscriptionSend.as_view(),
        name="subscription-send",
    ),
    url(
        r"^api/v1/subscriptions/(?P<subscription_id>.+)/resend$",
        views.SubscriptionResend.as_view(),
    ),
    url(r"^api/v1/subscriptions/request$", views.SubscriptionRequest.as_view()),
    url(r"^api/v1/user/token/$", views.UserView.as_view(), name="create-user-token"),
    url(r"^api/v1/runsendestimate$", views.DailyEstimateRun.as_view()),
]



import csv
import os

from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import BaseCommand, CommandError

from subscriptions.models import Subscription
from subscriptions.tasks import schedule_create


class Command(BaseCommand):
    help = (
        "Sometimes the task to create the schedule for a subscription "
        "fails. This takes a list of subscription ids and reruns the "
        "schedule creation task if they don't have a schedule."
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "--csv",
            type=str,
            help="CSV file containing the UUIDs of the subscriptions to check",
        )

    def handle(self, *args, **options):
        file_name = options["csv"]

        if not file_name:
            raise CommandError("--csv is a required parameter")
        if not os.path.isfile(file_name):
            raise CommandError("--csv must be a valid path to a file")

        self.stdout.write("Processing file ...")
        count = 0
        for sub_id in self.subscription_from_csv(file_name):
            try:
                subscription = Subscription.objects.get(pk=sub_id)
            except ObjectDoesNotExist:
                self.stdout.write("Subscription %s does not exist" % sub_id)
                continue
            except ValueError:
                self.stdout.write("%s is not a valid UUID" % sub_id)
                continue

            # Some schedules were created but failed before saving to the sub.
            # If it's been triggered it probably has a schedule so ignore
            if (
                (
                    subscription.metadata is not None
                    and "scheduler_schedule_id" in subscription.metadata
                )
                or subscription.next_sequence_number
                > subscription.initial_sequence_number
            ):
                self.stdout.write("Subscription %s already has schedule" % sub_id)
                continue
            else:
                result = schedule_create(sub_id)
                if not result:
                    self.stdout.write(
                        "Failed to create schedule for " "subscription %s" % sub_id
                    )
                else:
                    count += 1
                    self.stdout.write(
                        "Created schedule %s for subscription %s" % (result, sub_id)
                    )
        self.stdout.write("Created %s schedules" % count)

    def subscription_from_csv(self, csv_file):
        with open(csv_file) as csv_file:
            reader = csv.DictReader(csv_file)
            if not (set(["subscription"]) <= set(reader.fieldnames)):
                raise CommandError("CSV file must contain subscription.")
            for data in reader:
                yield data["subscription"]



from django.core.management.base import BaseCommand

from contentstore.models import Schedule
from contentstore.tasks import sync_schedule


class Command(BaseCommand):
    help = "Synchronises all schedules with the scheduler"

    def handle(self, *args, **kwargs):
        count = 0
        self.stdout.write("Fetching schedules...")

        for schedule in Schedule.objects.all().iterator():
            self.stdout.write("Synchronising schedule {}".format(schedule.id))
            sync_schedule(str(schedule.id))
            count += 1

        self.stdout.write(
            self.style.SUCCESS("Synchronised {} schedule/s".format(count))
        )



import django_filters
from django.contrib.auth.models import User
from django.contrib.postgres.fields import JSONField
from django_filters import rest_framework as filters
from rest_framework import mixins, status, viewsets
from rest_framework.authtoken.models import Token
from rest_framework.decorators import action
from rest_framework.pagination import CursorPagination
from rest_framework.permissions import (
    DjangoModelPermissions,
    IsAdminUser,
    IsAuthenticated,
)
from rest_framework.response import Response
from rest_framework.views import APIView

from seed_stage_based_messaging.utils import get_available_metrics

from .models import BehindSubscription, Subscription, SubscriptionSendFailure
from .serializers import (
    BehindSubscriptionSerializer,
    CreateUserSerializer,
    SubscriptionSendFailureSerializer,
    SubscriptionSerializer,
)
from .tasks import (
    find_behind_subscriptions,
    fire_daily_send_estimate,
    requeue_failed_tasks,
    schedule_disable,
    scheduled_metrics,
    store_resend_request,
)


class CreatedAtCursorPagination(CursorPagination):
    ordering = "-created_at"


class IdCursorPagination(CursorPagination):
    ordering = "-id"


class SubscriptionFilter(filters.FilterSet):
    created_after = django_filters.IsoDateTimeFilter(
        field_name="created_at", lookup_expr="gte"
    )
    created_before = django_filters.IsoDateTimeFilter(
        field_name="created_at", lookup_expr="lte"
    )
    metadata_has_key = django_filters.CharFilter(
        field_name="metadata", lookup_expr="has_key"
    )
    metadata_not_has_key = django_filters.CharFilter(
        field_name="metadata", lookup_expr="has_key", exclude=True
    )
    messageset_contains = django_filters.CharFilter(
        field_name="messageset__short_name", lookup_expr="contains"
    )

    class Meta:
        model = Subscription
        exclude = ()
        filter_overrides = {JSONField: {"filter_class": django_filters.CharFilter}}


class SubscriptionViewSet(viewsets.ModelViewSet):

    """ API endpoint that allows Subscription models to be viewed or edited.
    """

    permission_classes = (IsAuthenticated,)
    queryset = Subscription.objects.all().select_related("messageset")
    serializer_class = SubscriptionSerializer
    filterset_class = SubscriptionFilter
    pagination_class = CreatedAtCursorPagination


class SubscriptionSend(APIView):

    """ Triggers a send for the next subscription message
    """

    permission_classes = (IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        """ Validates subscription data before creating Outbound message
        """
        schedule_disable.delay(kwargs["subscription_id"])
        return Response({"accepted": True}, status=201)


class SubscriptionResend(APIView):

    """ Triggers a re-send for the current subscription message
    """

    permission_classes = (IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        """ Validates subscription data before creating Outbound message
        """
        # Look up subscriber
        subscription_id = kwargs["subscription_id"]
        if Subscription.objects.filter(id=subscription_id).exists():
            status = 202
            accepted = {"accepted": True}
            store_resend_request.apply_async(args=[subscription_id])
        else:
            status = 400
            accepted = {
                "accepted": False,
                "reason": "Cannot find subscription with ID {}".format(subscription_id),
            }
        return Response(accepted, status=status)


class SubscriptionRequest(APIView):

    """ Webhook listener for registrations now needing a subscription
    """

    permission_classes = (IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        """ Validates subscription data before creating Subscription message
        """
        # Ensure that we check for the 'data' key in the request object before
        # attempting to reference it
        if "data" in request.data:
            # This is a workaround for JSONField not liking blank/null refs
            if "metadata" not in request.data["data"]:
                request.data["data"]["metadata"] = {}

            if "initial_sequence_number" not in request.data["data"]:
                request.data["data"]["initial_sequence_number"] = request.data[
                    "data"
                ].get("next_sequence_number")

            subscription = SubscriptionSerializer(data=request.data["data"])
            if subscription.is_valid():
                subscription.save()
                # Return
                status = 201
                accepted = {"accepted": True}
                return Response(accepted, status=status)
            else:
                status = 400
                return Response(subscription.errors, status=status)
        else:
            status = 400
            message = {"data": ["This field is required."]}
            return Response(message, status=status)


class UserView(APIView):
    """ API endpoint that allows users creation and returns their token.
    Only admin users can do this to avoid permissions escalation.
    """

    permission_classes = (IsAdminUser,)

    def post(self, request):
        """Create a user and token, given an email. If user exists just
        provide the token."""
        serializer = CreateUserSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        email = serializer.validated_data.get("email")
        try:
            user = User.objects.get(username=email)
        except User.DoesNotExist:
            user = User.objects.create_user(email, email=email)
        token, created = Token.objects.get_or_create(user=user)

        return Response(status=status.HTTP_201_CREATED, data={"token": token.key})


class MetricsView(APIView):

    """ Metrics Interaction
        GET - returns list of all available metrics on the service
        POST - starts up the task that fires all the scheduled metrics
    """

    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        status = 200
        resp = {"metrics_available": get_available_metrics()}
        return Response(resp, status=status)

    def post(self, request, *args, **kwargs):
        status = 201
        scheduled_metrics.apply_async()
        resp = {"scheduled_metrics_initiated": True}
        return Response(resp, status=status)


class HealthcheckView(APIView):

    """ Healthcheck Interaction
        GET - returns service up - getting auth'd requires DB
    """

    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        import seed_stage_based_messaging
        import django
        import rest_framework

        status = 200
        resp = {
            "up": True,
            "result": {
                "database": "Accessible",
                "version": seed_stage_based_messaging.__version__,
                "libraries": {
                    "django": django.__version__,
                    "djangorestframework": rest_framework.__version__,
                },
            },
        }
        return Response(resp, status=status)


class FailedTaskViewSet(
    mixins.ListModelMixin, mixins.RetrieveModelMixin, viewsets.GenericViewSet
):
    permission_classes = (IsAuthenticated,)
    queryset = SubscriptionSendFailure.objects.all()
    serializer_class = SubscriptionSendFailureSerializer
    pagination_class = IdCursorPagination

    def create(self, request):
        status = 201
        resp = {"requeued_failed_tasks": True}
        requeue_failed_tasks.delay()
        return Response(resp, status=status)


class DailyEstimateRun(APIView):

    """ Triggers a daily send estimation
    """

    permission_classes = (IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        task_id = fire_daily_send_estimate.apply_async()
        accepted = {"accepted": True, "task_id": str(task_id)}
        return Response(accepted, status=202)


class DjangoModelViewPermissions(DjangoModelPermissions):
    """
    Django model permissions for viewing
    """

    perms_map = {"GET": ["%(app_label)s.view_%(model_name)s"]}


class CanFindBehindSubscriptionsPermission(DjangoModelPermissions):
    """
    Django model permissions for can_find_behind_subscriptions permission
    """

    perms_map = {"POST": ["%(app_label)s.can_find_behind_subscriptions"]}


class BehindSubscriptionViewSet(viewsets.ReadOnlyModelViewSet):
    """
    Viewset for querying BehindSubscriptions, and triggering the task that
    creates them.
    """

    queryset = BehindSubscription.objects.select_related(
        "current_messageset", "expected_messageset"
    ).all()
    serializer_class = BehindSubscriptionSerializer
    permission_classes = (DjangoModelViewPermissions,)
    pagination_class = CreatedAtCursorPagination

    @action(
        detail=False,
        methods=["post"],
        permission_classes=(CanFindBehindSubscriptionsPermission,),
    )
    def find_behind_subscriptions(self, request):
        """
        Starts a celery task that looks through active subscriptions to find
        and subscriptions that are behind where they should be, and adds a
        BehindSubscription for them.
        """

        task_id = find_behind_subscriptions.delay()
        return Response(
            {"accepted": True, "task_id": str(task_id)}, status=status.HTTP_202_ACCEPTED
        )



from django.core import exceptions
from django.core.management.base import BaseCommand
from django.db.models import Count
from django.utils.six.moves import input

from contentstore.models import MessageSet
from subscriptions.models import Subscription


def list_validator(valid_inputs):
    def validate(value):
        if value in valid_inputs:
            return value
        else:
            msg = "Selelection must be one of the following: {0}".format(
                str(valid_inputs)
            )
            raise exceptions.ValidationError(msg, code="invalid_choice")

    return validate


class Command(BaseCommand):
    help = "Resets subscriptions status to ready."

    def get_option(self, value, option_format="[{0:>3}]"):
        return option_format.format(value)

    def handle(self, *args, **options):
        selection = None
        confirm = None

        stuck_subscriptions = Subscription.objects.filter(process_status=1)
        stuck_totals = (
            stuck_subscriptions.values("messageset")
            .annotate(total=Count("messageset"))
            .order_by("messageset")
        )

        self.stdout.write("Totals of stuck Subscriptions per MessageSet:")
        for total in stuck_totals:
            ms = MessageSet.objects.get(pk=total["messageset"])
            opt = self.get_option(ms.pk)
            line = "{opt} {name:<40} Total: {total:>6}".format(
                opt=opt, name=ms.short_name, total=total["total"]
            )
            self.stdout.write(line)

        line = "{opt} {name:<40} Total: {total:>6}".format(
            opt=self.get_option("all"),
            name="All MessageSets",
            total=stuck_subscriptions.count(),
        )
        self.stdout.write(line)
        line = "{opt} Do nothing and quit".format(
            opt=self.get_option("q"),
            name="All MessageSets",
            total=stuck_subscriptions.count(),
        )
        self.stdout.write(line)

        msg = (
            "Please type the ID of the MessageSet you'd like to reset "
            "subscriptions for: "
        )
        valid_inputs = [str(o["messageset"]) for o in stuck_totals]
        valid_inputs.extend(["all", "q"])
        while selection is None:
            selection = self.get_input_data(msg, list_validator(valid_inputs))
            if selection == "q":
                self.stdout.write("Quiting.")
                return

        if selection == "all":
            records = stuck_subscriptions.all()
        else:
            records = stuck_subscriptions.filter(messageset=selection)

        msg = "Do you wish to update these subscriptions to " "ready status? [y/n] "
        while confirm is None:
            confirm = self.get_input_data(msg, list_validator(["y", "n"]))

        if confirm:
            rows_updated = records.update(process_status=0)
            msg = "Updated {0} rows to ready status".format(rows_updated)
            self.stdout.write(self.style.SUCCESS(msg))
        else:
            self.stdout.write("Ok, doing nothing.")

    def get_input_data(self, message, validator=None):
        raw_value = input(message)
        if validator is not None:
            try:
                val = validator(raw_value)
            except exceptions.ValidationError as e:
                self.stderr.write("Error: %s" % "; ".join(e.messages))
                val = None
        return val



"""
Unit tests for contentstore signals
"""
from unittest.mock import patch

from django.test import TestCase

from contentstore.models import Schedule
from seed_stage_based_messaging import test_utils as utils


class ScheduleSignalsTests(TestCase):
    """
    Tests for signals relating to the Schedule model
    """

    @patch("contentstore.tasks.SyncSchedule.scheduler")
    def test_schedule_created(self, scheduler):
        """
        When a schedule is created, then a corresponding schedule should be
        created in the scheduler service, and the ID of that schedule saved
        onto the model.
        """
        scheduler.create_schedule.return_value = {
            "id": "460fa767-2540-4ab7-bfce-e2cf54efdeb2"
        }
        schedule = Schedule.objects.create()

        scheduler.create_schedule.assert_called_once_with(schedule.scheduler_format)
        schedule.refresh_from_db()
        self.assertEqual(
            str(schedule.scheduler_schedule_id), "460fa767-2540-4ab7-bfce-e2cf54efdeb2"
        )

    @patch("contentstore.tasks.SyncSchedule.scheduler")
    def test_schedule_modified(self, scheduler):
        """
        When a schedule is modified, then the corresponding schedule should be
        updated in the scheduler service.
        """
        scheduler.create_schedule.return_value = {
            "id": "460fa767-2540-4ab7-bfce-e2cf54efdeb2"
        }
        schedule = Schedule.objects.create()
        schedule.refresh_from_db()
        schedule.day_of_week = "1"
        schedule.save()

        scheduler.update_schedule.assert_called_with(
            "460fa767-2540-4ab7-bfce-e2cf54efdeb2", schedule.scheduler_format
        )

    @patch("contentstore.tasks.DeactivateSchedule.scheduler")
    def test_schedule_deleted(self, scheduler):
        """
        When a schedule is deleted, then the corresponding schedule should be
        deactivated in the scheduler service.
        """
        utils.disable_signals()
        schedule = Schedule.objects.create()
        schedule.scheduler_schedule_id = "460fa767-2540-4ab7-bfce-e2cf54efdeb2"
        schedule.save()
        utils.enable_signals()
        schedule.delete()

        scheduler.update_schedule.assert_called_with(
            "460fa767-2540-4ab7-bfce-e2cf54efdeb2", {"active": False}
        )



from django.conf.urls import include, url
from rest_framework import routers

from . import views

router = routers.DefaultRouter()
router.register(r"schedule", views.ScheduleViewSet)
router.register(r"messageset", views.MessageSetViewSet)
router.register(r"message", views.MessageViewSet)
router.register(r"binarycontent", views.BinaryContentViewSet)

# Wire up our API using automatic URL routing.
# Additionally, we include login URLs for the browseable API.
urlpatterns = [
    url(r"^api/v1/", include(router.urls)),
    url(
        r"^api/v1/message/(?P<pk>.+)/content$",
        views.MessagesContentView.as_view({"get": "retrieve"}),
    ),
    url(
        r"^api/v1/messageset/(?P<pk>.+)/messages$",
        views.MessagesetMessagesContentView.as_view({"get": "retrieve"}),
    ),
    url(r"^api/v1/messageset_languages/$", views.MessagesetLanguageView.as_view()),
    url(r"^api/v1/sync_audio_files/$", views.SyncAudioFilesView.as_view()),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-02-13 09:59
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("subscriptions", "0004_subscription_initial_sequence_number")]

    operations = [
        migrations.CreateModel(
            name="SubscriptionSendFailure",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("task_id", models.UUIDField()),
                ("initiated_at", models.DateTimeField()),
                ("reason", models.TextField()),
                (
                    "subscription",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="subscriptions.Subscription",
                    ),
                ),
            ],
        )
    ]



from rest_framework import serializers

from .models import BehindSubscription, Subscription, SubscriptionSendFailure


class CreateUserSerializer(serializers.Serializer):
    email = serializers.EmailField()


class SubscriptionSerializer(serializers.ModelSerializer):
    messageset_label = serializers.SlugRelatedField(
        source="messageset", slug_field="label", read_only=True
    )

    class Meta:
        model = Subscription
        read_only_fields = ("url", "id", "created_at", "updated_at")
        fields = (
            "url",
            "id",
            "version",
            "identity",
            "messageset",
            "next_sequence_number",
            "lang",
            "active",
            "completed",
            "schedule",
            "process_status",
            "metadata",
            "created_at",
            "updated_at",
            "initial_sequence_number",
            "messageset_label",
        )


class SubscriptionSendFailureSerializer(serializers.HyperlinkedModelSerializer):
    subscription_id = serializers.PrimaryKeyRelatedField(read_only=True)

    class Meta:
        model = SubscriptionSendFailure
        fields = (
            "url",
            "id",
            "subscription",
            "subscription_id",
            "task_id",
            "initiated_at",
            "reason",
        )


class BehindSubscriptionSerializer(serializers.HyperlinkedModelSerializer):
    subscription_id = serializers.PrimaryKeyRelatedField(read_only=True)
    current_messageset_id = serializers.PrimaryKeyRelatedField(read_only=True)
    current_messageset_name = serializers.StringRelatedField(
        source="current_messageset"
    )
    expected_messageset_id = serializers.PrimaryKeyRelatedField(read_only=True)
    expected_messageset_name = serializers.StringRelatedField(
        source="expected_messageset"
    )

    class Meta:
        model = BehindSubscription
        fields = (
            "id",
            "subscription",
            "subscription_id",
            "messages_behind",
            "current_messageset",
            "current_messageset_id",
            "current_messageset_name",
            "current_sequence_number",
            "expected_messageset",
            "expected_messageset_id",
            "expected_messageset_name",
            "expected_sequence_number",
            "created_at",
        )



import json
from datetime import datetime

from django.core.management.base import BaseCommand
from django.utils import timezone

from subscriptions.models import Subscription
from subscriptions.tasks import send_next_message


class Command(BaseCommand):
    help = (
        "This command is used when the subscription has fallen behind "
        "schedule. Leave the action argument blank to see how many "
        "subscriptions are behind. Running the command with `--action "
        "send` will send a message to each subscription that is behind. "
        "Running the command with `--action fast_forward` will fast "
        "forward the subscriptions that are behind to the end_date. "
        "Running the command with `--action diff` will print out the "
        "differences that running the command would make."
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "--end_date",
            dest="end_date",
            default=datetime.now(),
            type=lambda today: datetime.strptime(today, "%Y%m%d"),
            help="""Fast forward subscription to end_date
                  By default it will use datetime.now() (format YYYYMMDD)""",
        )
        parser.add_argument(
            "--action",
            dest="action",
            default=False,
            help=(
                "Set to `send` to send next message or `fast_forward` to "
                "fast forward the subscription or `diff` to print out the "
                "changes that the command would make."
            ),
        )
        parser.add_argument(
            "--verbose",
            dest="verbose",
            default=False,
            help=("Print out some details on the relevant subscriptions."),
        )
        parser.add_argument(
            "--message-set",
            dest="message_set",
            default=None,
            type=int,
            help=(
                "Only apply the action to the subscriptions that are for "
                "the specified message set, defaults to all message sets."
            ),
        )
        parser.add_argument(
            "--messages-limit",
            dest="messages_limit",
            default=None,
            type=int,
            help=(
                "Only apply the action to subscriptions that are behind by"
                "the limit or less than the limit. Defaults to no limit."
            ),
        )

    def handle(self, *args, **options):
        action = options["action"]
        verbose = options["verbose"]
        end_date = options["end_date"]
        end_date = end_date.replace(tzinfo=timezone.utc)
        message_set = options["message_set"]
        messages_limit = options["messages_limit"]

        behind = 0
        forwards = 0
        sends = 0

        subscriptions = Subscription.objects.filter(active=True, process_status=0)
        if message_set is not None:
            subscriptions = subscriptions.filter(messageset__pk=message_set)

        for sub in subscriptions.iterator():
            messages_behind = sub.messages_behind()

            if messages_limit is not None and messages_behind > messages_limit:
                continue

            if messages_behind > 0:

                if verbose:
                    self.stdout.write("{}: {}".format(sub.id, messages_behind))

                if action == "fast_forward":
                    Subscription.fast_forward_lifecycle(sub, end_date)
                    forwards += 1
                elif action == "send":
                    send_next_message.apply_async(args=[str(sub.id)])
                    sends += 1
                elif action == "diff":
                    start_ms = sub.messageset.pk
                    start_nseq = sub.next_sequence_number
                    subs = Subscription.fast_forward_lifecycle(
                        sub, end_date, save=False
                    )
                    end_sub = subs[-1]
                    self.stdout.write(
                        json.dumps(
                            {
                                "language": sub.lang,
                                "identity": sub.identity,
                                "current_messageset_id": start_ms,
                                "current_sequence_number": start_nseq,
                                "expected_messageset_id": end_sub.messageset.pk,
                                "expected_sequence_number": end_sub.next_sequence_number,
                                "messages_behind": messages_behind,
                            }
                        )
                    )

                behind += 1

        self.stdout.write(
            "%s subscription%s behind schedule." % (behind, "" if behind == 1 else "s")
        )
        self.stdout.write(
            "%s subscription%s fast forwarded to end date."
            % (forwards, "" if forwards == 1 else "s")
        )
        self.stdout.write(
            "Message sent to %s subscription%s." % (sends, "" if sends == 1 else "s")
        )



from rest_framework import serializers

from .models import BinaryContent, Message, MessageSet, Schedule


class ScheduleSerializer(serializers.ModelSerializer):
    class Meta:
        model = Schedule
        fields = (
            "id",
            "minute",
            "hour",
            "day_of_week",
            "day_of_month",
            "month_of_year",
        )


class MessageSetSerializer(serializers.ModelSerializer):
    class Meta:
        model = MessageSet
        fields = (
            "id",
            "short_name",
            "label",
            "content_type",
            "notes",
            "next_set",
            "default_schedule",
            "created_at",
            "updated_at",
        )


class BinaryContentSerializer(serializers.ModelSerializer):
    class Meta:
        model = BinaryContent
        fields = ("id", "content")


class MessageSerializer(serializers.ModelSerializer):
    class Meta:
        model = Message
        fields = (
            "id",
            "messageset",
            "sequence_number",
            "lang",
            "metadata",
            "text_content",
            "binary_content",
            "created_at",
            "updated_at",
        )


class MessageListSerializer(serializers.ModelSerializer):

    """
        Only used for get views, because binary relational serializer is not
        something that works nicely with posts
    """

    binary_content = BinaryContentSerializer(many=False, read_only=True)

    class Meta:
        model = Message
        fields = (
            "id",
            "messageset",
            "sequence_number",
            "lang",
            "text_content",
            "binary_content",
            "created_at",
            "updated_at",
        )


class MessageSetMessagesSerializer(serializers.ModelSerializer):
    messages = MessageListSerializer(many=True, read_only=True)

    class Meta:
        model = MessageSet
        fields = (
            "id",
            "short_name",
            "notes",
            "next_set",
            "default_schedule",
            "messages",
            "created_at",
            "updated_at",
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-07-27 08:20
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("subscriptions", "0005_subscriptionsendfailure")]

    operations = [
        migrations.AlterField(
            model_name="subscription",
            name="identity",
            field=models.CharField(db_index=True, max_length=36),
        )
    ]



# Generated by Django 2.1.2 on 2018-11-26 11:11

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("contentstore", "0010_auto_20181126_1104"),
        ("subscriptions", "0008_resendrequest"),
    ]

    operations = [
        migrations.CreateModel(
            name="BehindSubscription",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "messages_behind",
                    models.IntegerField(
                        help_text="The number of messages the subscription is behind by"
                    ),
                ),
                (
                    "current_sequence_number",
                    models.IntegerField(
                        help_text="Which sequence in the messageset we are at"
                    ),
                ),
                (
                    "expected_sequence_number",
                    models.IntegerField(
                        help_text="Which sequence in the messageset we expect to be"
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                (
                    "current_messageset",
                    models.ForeignKey(
                        help_text="The message set the the subscription is on",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="contentstore.MessageSet",
                    ),
                ),
                (
                    "expected_messageset",
                    models.ForeignKey(
                        help_text="The messageset that the subscription should be on",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="contentstore.MessageSet",
                    ),
                ),
            ],
        ),
        migrations.AlterField(
            model_name="resendrequest",
            name="message",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="resend_requests",
                to="contentstore.Message",
            ),
        ),
        migrations.AlterField(
            model_name="subscription",
            name="created_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="subscriptions_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterField(
            model_name="subscription",
            name="messageset",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="subscriptions",
                to="contentstore.MessageSet",
            ),
        ),
        migrations.AlterField(
            model_name="subscription",
            name="schedule",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="subscriptions",
                to="contentstore.Schedule",
            ),
        ),
        migrations.AlterField(
            model_name="subscription",
            name="updated_by",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="subscriptions_updated",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="behindsubscription",
            name="subscription",
            field=models.ForeignKey(
                help_text="The subscription that is behind",
                on_delete=django.db.models.deletion.CASCADE,
                to="subscriptions.Subscription",
            ),
        ),
    ]
